# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import os
import sys
from time import (sleep, time)
from random import randrange
import math
import argparse
import traceback
import logging
from random import randrange

from pyocd.core.target import Target
from pyocd.coresight.cortex_m import CortexM
from pyocd.gdbserver.context_facade import GDBDebugContextFacade
from pyocd.core.helpers import ConnectHelper
from pyocd.utility.conversion import float32_to_u32, u32_to_float32
from pyocd.utility.mask import same
from pyocd.core import exceptions
from pyocd.core.memory_map import MemoryType
from pyocd.flash.file_programmer import FileProgrammer

from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    get_test_binary_path,
    )

TEST_COUNT = 20

class CortexTestResult(TestResult):
    METRICS = [
        "get_t_response",
        "step",
        "bp_add_remove",
        "get_reg_context",
        "set_reg_context",
        "run_halt",
        "gdb_step",
        ]
    
    def __init__(self):
        super(CortexTestResult, self).__init__(None, None, None)
        self.name = "cortex"
        self.times = {}
        for metric in self.METRICS:
            self.times[metric] = -1.0

class CortexTest(Test):
    def __init__(self):
        super(CortexTest, self).__init__("Cortex Test", cortex_test)

    def print_perf_info(self, result_list, output_file=None):
        result_list = filter(lambda x: isinstance(x, CortexTestResult), result_list)
        print("\n\n------ Cortex Test Performance ------", file=output_file)
        print("", file=output_file)
        for result in result_list:
            print("Target {}:".format(result.board), file=output_file)
            for metric in CortexTestResult.METRICS:
                print("    {:<20} = {: 8.3f} ms".format(metric, result.times[metric] * 1000), file=output_file)
        print("", file=output_file)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            result = CortexTestResult()
            result.passed = False
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def float_compare(f1, f2):
    return abs(f1 - f2) < 0.0001

def test_function(session, function):
    session.probe.flush()
    start = time()
    for i in range(0, TEST_COUNT):
        function()
        session.probe.flush()
    stop = time()
    return (stop - start) / float(TEST_COUNT)

def cortex_test(board_id):
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        target_type = board.target_type

        binary_file = get_test_binary_path(board.test_binary)

        test_params = get_target_test_params(session)
        test_clock = test_params['test_clock']
        addr_invalid = 0x3E000000 # Last 16MB of ARM SRAM region - typically empty
        expect_invalid_access_to_fail = test_params['error_on_invalid_access']

        memory_map = board.target.get_memory_map()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        rom_region = memory_map.get_boot_memory()

        addr = ram_region.start
        size = 0x502
        addr_bin = rom_region.start

        target = board.target
        probe = session.probe

        probe.set_clock(test_clock)

        test_pass_count = 0
        test_count = 0
        result = CortexTestResult()

        debugContext = target.get_target_context()
        gdbFacade = GDBDebugContextFacade(debugContext)

        print("\n\n----- FLASH NEW BINARY BEFORE TEST -----")
        FileProgrammer(session).program(binary_file, base_address=addr_bin)
        # Let the target run for a bit so it
        # can initialize the watchdog if it needs to
        target.resume()
        sleep(0.2)
        target.halt()

        print("PROGRAMMING COMPLETE")


        print("\n\n----- TESTING CORTEX-M PERFORMANCE -----")
        test_time = test_function(session, gdbFacade.get_t_response)
        result.times["get_t_response"] = test_time
        print("Function get_t_response time:    %f" % test_time)

        # Step
        test_time = test_function(session, target.step)
        result.times["step"] = test_time
        print("Function step time:              %f" % test_time)

        # Breakpoint
        def set_remove_breakpoint():
            target.set_breakpoint(0)
            target.remove_breakpoint(0)
        test_time = test_function(session, set_remove_breakpoint)
        result.times["bp_add_remove"] = test_time
        print("Add and remove breakpoint:       %f" % test_time)

        # get_register_context
        test_time = test_function(session, gdbFacade.get_register_context)
        result.times["get_reg_context"] = test_time
        print("Function get_register_context:   %f" % test_time)

        # set_register_context
        context = gdbFacade.get_register_context()
        def set_register_context():
            gdbFacade.set_register_context(context)
        test_time = test_function(session, set_register_context)
        result.times["set_reg_context"] = test_time
        print("Function set_register_context:   %f" % test_time)

        # Run / Halt
        def run_halt():
            target.resume()
            target.halt()
        test_time = test_function(session, run_halt)
        result.times["run_halt"] = test_time
        print("Resume and halt:                 %f" % test_time)

        # GDB stepping
        def simulate_step():
            target.step()
            gdbFacade.get_t_response()
            target.set_breakpoint(0)
            target.resume()
            target.halt()
            gdbFacade.get_t_response()
            target.remove_breakpoint(0)
        test_time = test_function(session, simulate_step)
        result.times["gdb_step"] = test_time
        print("Simulated GDB step:              %f" % test_time)

        # Test passes if there are no exceptions
        test_pass_count += 1
        test_count += 1
        print("TEST PASSED")

        print("\n\n------ Testing Reset Types ------")
        def reset_methods(fnc):
            print("Hardware reset")
            fnc(reset_type=Target.ResetType.HW)
            print("Hardware reset (default=HW)")
            target.selected_core.default_reset_type = Target.ResetType.HW
            fnc(reset_type=None)
            print("Software reset (default=SYSRESETREQ)")
            target.selected_core.default_reset_type = Target.ResetType.SW_SYSRESETREQ
            fnc(reset_type=None)
            print("Software reset (default=VECTRESET)")
            target.selected_core.default_reset_type = Target.ResetType.SW_VECTRESET
            fnc(reset_type=None)
            print("Software reset (default=emulated)")
            target.selected_core.default_reset_type = Target.ResetType.SW_EMULATED
            fnc(reset_type=None)
            
            print("(Default) Software reset (SYSRESETREQ)")
            target.selected_core.default_software_reset_type = Target.ResetType.SW_SYSRESETREQ
            fnc(reset_type=Target.ResetType.SW)
            print("(Default) Software reset (VECTRESET)")
            target.selected_core.default_software_reset_type = Target.ResetType.SW_VECTRESET
            fnc(reset_type=Target.ResetType.SW)
            print("(Default) Software reset (emulated)")
            target.selected_core.default_software_reset_type = Target.ResetType.SW_EMULATED
            fnc(reset_type=Target.ResetType.SW)
            
            print("Software reset (option=default)")
            target.selected_core.default_reset_type = Target.ResetType.SW
            target.selected_core.default_software_reset_type = Target.ResetType.SW_SYSRESETREQ
            session.options['reset_type'] = 'default'
            fnc(reset_type=None)
            print("Software reset (option=hw)")
            session.options['reset_type'] = 'hw'
            fnc(reset_type=None)
            print("Software reset (option=sw)")
            session.options['reset_type'] = 'sw'
            fnc(reset_type=None)
            print("Software reset (option=sw_sysresetreq)")
            session.options['reset_type'] = 'sw_sysresetreq'
            fnc(reset_type=None)
            print("Software reset (option=sw_vectreset)")
            session.options['reset_type'] = 'sw_vectreset'
            fnc(reset_type=None)
            print("Software reset (option=sw_emulated)")
            session.options['reset_type'] = 'sw_emulated'
            fnc(reset_type=None)

        reset_methods(target.reset)
        
        # Test passes if there are no exceptions
        test_pass_count += 1
        test_count += 1
        print("TEST PASSED")


        print("\n\n------ Testing Reset Halt ------")
        reset_methods(target.reset_and_halt)

        # Test passes if there are no exceptions
        test_pass_count += 1
        test_count += 1
        print("TEST PASSED")


        print("\n\n------ Testing Register Read/Write ------")
        print("Reading r0")
        val = target.read_core_register('r0')
        origR0 = val
        rawVal = target.read_core_register_raw('r0')
        test_count += 1
        if val == rawVal:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")

        print("Writing r0")
        target.write_core_register('r0', 0x12345678)
        val = target.read_core_register('r0')
        rawVal = target.read_core_register_raw('r0')
        test_count += 1
        if val == 0x12345678 and rawVal == 0x12345678:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")

        print("Raw writing r0")
        target.write_core_register_raw('r0', 0x87654321)
        val = target.read_core_register('r0')
        rawVal = target.read_core_register_raw('r0')
        test_count += 1
        if val == 0x87654321 and rawVal == 0x87654321:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")

        print("Read/write r0, r1, r2, r3")
        origRegs = target.read_core_registers_raw(['r0', 'r1', 'r2', 'r3'])
        target.write_core_registers_raw(['r0', 'r1', 'r2', 'r3'], [1, 2, 3, 4])
        vals = target.read_core_registers_raw(['r0', 'r1', 'r2', 'r3'])
        passed = vals[0] == 1 and vals[1] == 2 and vals[2] == 3 and vals[3] == 4
        test_count += 1
        if passed:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")
            
        # Restore regs
        origRegs[0] = origR0
        target.write_core_registers_raw(['r0', 'r1', 'r2', 'r3'], origRegs)
        
        print("Verify exception is raised while core is running")
        target.resume()
        try:
            val = target.read_core_register('r0')
        except exceptions.CoreRegisterAccessError:
            passed = True
        else:
            passed = False
        test_count += 1
        if passed:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")
        
        print("Verify failure to write core register while running raises exception")
        try:
            target.write_core_register('r0', 0x1234)
        except exceptions.CoreRegisterAccessError:
            passed = True
        else:
            passed = False
        test_count += 1
        if passed:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")
        
        # Resume execution.
        target.halt()

        if target.selected_core.has_fpu:
            print("Reading s0")
            val = target.read_core_register('s0')
            rawVal = target.read_core_register_raw('s0')
            origRawS0 = rawVal
            passed = isinstance(val, float) and isinstance(rawVal, int) \
                        and float32_to_u32(val) == rawVal
            test_count += 1
            if passed:
                test_pass_count += 1
                print("TEST PASSED")
            else:
                print("TEST FAILED")

            print("Writing s0")
            target.write_core_register('s0', math.pi)
            val = target.read_core_register('s0')
            rawVal = target.read_core_register_raw('s0')
            passed = float_compare(val, math.pi) and float_compare(u32_to_float32(rawVal), math.pi)
            test_count += 1
            if passed:
                test_pass_count += 1
                print("TEST PASSED")
            else:
                print("TEST FAILED (%f==%f, 0x%08x->%f)" % (val, math.pi, rawVal, u32_to_float32(rawVal)))

            print("Raw writing s0")
            x = float32_to_u32(32.768)
            target.write_core_register_raw('s0', x)
            val = target.read_core_register('s0')
            passed = float_compare(val, 32.768)
            test_count += 1
            if passed:
                test_pass_count += 1
                print("TEST PASSED")
            else:
                print("TEST FAILED (%f==%f)" % (val, 32.768))

            print("Read/write s0, s1")
            _1p1 = float32_to_u32(1.1)
            _2p2 = float32_to_u32(2.2)
            origRegs = target.read_core_registers_raw(['s0', 's1'])
            target.write_core_registers_raw(['s0', 's1'], [_1p1, _2p2])
            vals = target.read_core_registers_raw(['s0', 's1'])
            s0 = target.read_core_register('s0')
            s1 = target.read_core_register('s1')
            passed = vals[0] == _1p1 and float_compare(s0, 1.1) \
                        and vals[1] == _2p2 and float_compare(s1, 2.2)
            test_count += 1
            if passed:
                test_pass_count += 1
                print("TEST PASSED")
            else:
                print("TEST FAILED (0x%08x==0x%08x, %f==%f, 0x%08x==0x%08x, %f==%f)" \
                    % (vals[0], _1p1, s0, 1.1, vals[1], _2p2, s1, 2.2))
            
            # Restore s0
            origRegs[0] = origRawS0
            target.write_core_registers_raw(['s0', 's1'], origRegs)
        
        print("Verify that all listed core registers can be accessed")
        reg_count = 0
        passed_reg_count = 0
        for r in target.selected_core.core_registers.as_set:
            try:
                reg_count += 1
                val = target.read_core_register(r.name)
                target.write_core_register(r.name, val)
                passed_reg_count += 1
            except exceptions.CoreRegisterAccessError:
                pass
        test_count += 1
        if passed_reg_count == reg_count:
            test_pass_count += 1
            print("TEST PASSED (%i registers)" % reg_count)
        else:
            print("TEST FAILED (%i registers, %i failed)" % (reg_count, reg_count - passed_reg_count))

        print("\n\n------ Testing Invalid Memory Access Recovery ------")
        memory_access_pass = True
        try:
            print("reading 0x1000 bytes at invalid address 0x%08x" % addr_invalid)
            target.read_memory_block8(addr_invalid, 0x1000)
            target.flush()
            # If no exception is thrown the tests fails except on nrf51 where invalid addresses read as 0
            if expect_invalid_access_to_fail:
                print("  failed to get expected fault")
                memory_access_pass = False
            else:
                print("  no fault as expected")
        except exceptions.TransferFaultError as exc:
            print("  got expected error: " + str(exc))

        try:
            print("reading 0x1000 bytes at invalid address 0x%08x" % (addr_invalid + 1))
            target.read_memory_block8(addr_invalid + 1, 0x1000)
            target.flush()
            # If no exception is thrown the tests fails except on nrf51 where invalid addresses read as 0
            if expect_invalid_access_to_fail:
                print("  failed to get expected fault")
                memory_access_pass = False
            else:
                print("  no fault as expected")
        except exceptions.TransferFaultError as exc:
            print("  got expected error: " + str(exc))

        data = [0x00] * 0x1000
        try:
            print("writing 0x%08x bytes at invalid address 0x%08x" % (len(data), addr_invalid))
            target.write_memory_block8(addr_invalid, data)
            target.flush()
            # If no exception is thrown the tests fails except on nrf51 where invalid addresses read as 0
            if expect_invalid_access_to_fail:
                print("  failed to get expected fault!")
                memory_access_pass = False
            else:
                print("  no fault as expected")
        except exceptions.TransferFaultError as exc:
            print("  got expected error: " + str(exc))

        data = [0x00] * 0x1000
        try:
            print("writing 0x%08x bytes at invalid address 0x%08x" % (len(data), addr_invalid + 1))
            target.write_memory_block8(addr_invalid + 1, data)
            target.flush()
            # If no exception is thrown the tests fails except on nrf51 where invalid addresses read as 0
            if expect_invalid_access_to_fail:
                print("  failed to get expected fault!")
                memory_access_pass = False
            else:
                print("  no fault as expected")
        except exceptions.TransferFaultError as exc:
            print("  got expected error: " + str(exc))

        data = [randrange(0, 255) for x in range(size)]
        print("r/w 0x%08x bytes at 0x%08x" % (size, addr))
        target.write_memory_block8(addr, data)
        block = target.read_memory_block8(addr, size)
        if same(data, block):
            print("  Aligned access pass")
        else:
            print("  Memory read does not match memory written")
            memory_access_pass = False

        data = [randrange(0, 255) for x in range(size)]
        print("r/w 0x%08x bytes at 0x%08x" % (size, addr + 1))
        target.write_memory_block8(addr + 1, data)
        block = target.read_memory_block8(addr + 1, size)
        if same(data, block):
            print("  Unaligned access pass")
        else:
            print("  Unaligned memory read does not match memory written")
            memory_access_pass = False

        test_count += 1
        if memory_access_pass:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")

        print("\n\n------ Testing Software Breakpoints ------")
        test_passed = True
        orig8x2 = target.read_memory_block8(addr, 2)
        orig8 = target.read8(addr)
        orig16 = target.read16(addr & ~1)
        orig32 = target.read32(addr & ~3)
        origAligned32 = target.read_memory_block32(addr & ~3, 1)

        def test_filters():
            test_passed = True
            filtered = target.read_memory_block8(addr, 2)
            if same(orig8x2, filtered):
                print("2 byte unaligned passed")
            else:
                print("2 byte unaligned failed (read %x-%x, expected %x-%x)" % (filtered[0], filtered[1], orig8x2[0], orig8x2[1]))
                test_passed = False

            for now in (True, False):
                filtered = target.read8(addr, now)
                if not now:
                    filtered = filtered()
                if filtered == orig8:
                    print("8-bit passed [now=%s]" % now)
                else:
                    print("8-bit failed [now=%s] (read %x, expected %x)" % (now, filtered, orig8))
                    test_passed = False

                filtered = target.read16(addr & ~1, now)
                if not now:
                    filtered = filtered()
                if filtered == orig16:
                    print("16-bit passed [now=%s]" % now)
                else:
                    print("16-bit failed [now=%s] (read %x, expected %x)" % (now, filtered, orig16))
                    test_passed = False

                filtered = target.read32(addr & ~3, now)
                if not now:
                    filtered = filtered()
                if filtered == orig32:
                    print("32-bit passed [now=%s]" % now)
                else:
                    print("32-bit failed [now=%s] (read %x, expected %x)" % (now, filtered, orig32))
                    test_passed = False

            filtered = target.read_memory_block32(addr & ~3, 1)
            if same(filtered, origAligned32):
                print("32-bit aligned passed")
            else:
                print("32-bit aligned failed (read %x, expected %x)" % (filtered[0], origAligned32[0]))
                test_passed = False
            return test_passed

        print("Installed software breakpoint at 0x%08x" % addr)
        target.set_breakpoint(addr, Target.BreakpointType.SW)
        test_passed = test_filters() and test_passed

        print("Removed software breakpoint")
        target.remove_breakpoint(addr)
        test_passed = test_filters() and test_passed

        test_count += 1
        if test_passed:
            test_pass_count += 1
            print("TEST PASSED")
        else:
            print("TEST FAILED")

        print("\nTest Summary:")
        print("Pass count %i of %i tests" % (test_pass_count, test_count))
        if test_pass_count == test_count:
            print("CORTEX TEST PASSED")
        else:
            print("CORTEX TEST FAILED")

        target.reset()

        result.passed = test_count == test_pass_count
        return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD cpu test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    cortex_test(None)



# pyOCD debugger
# Copyright (c) 2016 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .provider import TargetThread
from ..core import exceptions

LOG = logging.getLogger(__name__)

## Mask on EXC_RETURN indicating whether space for FP registers is allocated
# on the frame. The bit is 0 if the frame is extended.
EXC_RETURN_EXT_FRAME_MASK = (1 << 4)

def read_c_string(context, ptr):
    """! @brief Reads a null-terminated C string from the target."""
    if ptr == 0:
        return ""

    s = ""
    done = False
    count = 0
    badCount = 0
    try:
        while not done and count < 256:
            data = context.read_memory_block8(ptr, 16)
            ptr += 16
            count += 16

            for c in data:
                if c == 0:
                    done = True
                    break
                elif c > 127:
                    # Replace non-ASCII characters. If there is a run of invalid characters longer
                    # than 4, then terminate the string early.
                    badCount += 1
                    if badCount > 4:
                        done = True
                        break
                    s += '?'
                else:
                    s += chr(c)
                    badCount = 0
    except exceptions.TransferError:
        LOG.debug("TransferError while trying to read 16 bytes at 0x%08x", ptr)

    return s

class HandlerModeThread(TargetThread):
    """! @brief Class representing the handler mode."""

    UNIQUE_ID = 2
    
    def __init__(self, targetContext, provider):
        super(HandlerModeThread, self).__init__()
        self._target_context = targetContext
        self._provider = provider

    def get_stack_pointer(self):
        return self._target_context.read_core_register('msp')

    @property
    def priority(self):
        return 0

    @property
    def unique_id(self):
        return self.UNIQUE_ID

    @property
    def name(self):
        return "Handler mode"

    @property
    def description(self):
        ipsr = self._target_context.read_core_register('ipsr');
        return self._target_context.core.exception_number_to_name(ipsr)

    @property
    def is_current(self):
        return self._target_context.read_core_register('ipsr') > 0

    @property
    def context(self):
        return self._target_context

    def __str__(self):
        return "<HandlerModeThread@0x%08x>" % (id(self))

    def __repr__(self):
        return str(self)






# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
from typing import List

from .base import SubcommandBase
from ..commands.commander import PyOCDCommander
from ..utility.cmdline import (
    flatten_args,
    split_command_line,
)

class CommanderSubcommand(SubcommandBase):
    """! @brief `pyocd commander` subcommand."""
    
    NAMES = ['commander', 'cmd']
    HELP = "Interactive command console."
    DEFAULT_LOG_LEVEL = logging.WARNING

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        commander_parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        commander_options = commander_parser.add_argument_group("commander options")
        commander_options.add_argument("-H", "--halt", action="store_true", default=None,
            help="Halt core upon connect. (Deprecated, see --connect.)")
        commander_options.add_argument("-N", "--no-init", action="store_true",
            help="Do not init debug system.")
        commander_options.add_argument("--elf", metavar="PATH",
            help="Optionally specify ELF file being debugged.")
        commander_options.add_argument("-c", "--command", dest="commands", metavar="CMD", action='append', nargs='+',
            help="Run commands.")
        
        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, commander_parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'commander' subcommand."""
        # Flatten commands list then extract primary command and its arguments.
        if self._args.commands is not None:
            cmds = []
            for cmd in self._args.commands:
                cmds.append(flatten_args(split_command_line(arg) for arg in cmd))
        else:
            cmds = None

        # Enter REPL.
        PyOCDCommander(self._args, cmds).run()

        return 0





# pyOCD debugger
# Copyright (c) 2013-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
import logging
from enum import Enum

from ..core.target import Target
from ..core.exceptions import (FlashFailure, FlashEraseFailure, FlashProgramFailure)
from ..utility.mask import (align_down, msb)
from ..utility.timeout import Timeout
from .builder import FlashBuilder

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

# Program to compute the CRC of sectors.  This works on cortex-m processors.
# Code is relocatable and only needs to be on a 4 byte boundary.
# 200 bytes of executable data below + 1024 byte crc table = 1224 bytes
# Usage requirements:
# -In memory reserve 0x600 for code & table
# -Make sure data buffer is big enough to hold 4 bytes for each page that could be checked (ie.  >= num pages * 4)
_ANALYZER_CODE = (
    0x2780b5f0, 0x25004684, 0x4e2b2401, 0x447e4a2b, 0x0023007f, 0x425b402b, 0x40130868, 0x08584043,
    0x425b4023, 0x40584013, 0x40200843, 0x40104240, 0x08434058, 0x42404020, 0x40584010, 0x40200843,
    0x40104240, 0x08434058, 0x42404020, 0x40584010, 0x40200843, 0x40104240, 0x08584043, 0x425b4023,
    0x40434013, 0xc6083501, 0xd1d242bd, 0xd01f2900, 0x46602301, 0x469c25ff, 0x00894e11, 0x447e1841,
    0x88034667, 0x409f8844, 0x2f00409c, 0x2201d012, 0x4252193f, 0x34017823, 0x402b4053, 0x599b009b,
    0x405a0a12, 0xd1f542bc, 0xc00443d2, 0xd1e74281, 0xbdf02000, 0xe7f82200, 0x000000b2, 0xedb88320,
    0x00000042, 
    )

@dataclass
class SectorInfo:
    """@brief Info about an erase sector."""
    base_addr: int           # Start address of this sector
    erase_weight: int        # Time it takes to erase a page
    size: int                # Size of sector

@dataclass
class PageInfo:
    """@brief Info about a program page."""
    base_addr: int           # Start address of this page
    program_weight: int      # Time it takes to program a page (Not including data transfer time)
    size: int                # Size of page

@dataclass
class FlashInfo:
    """@brief Info about the entire flash region."""
    rom_start: int           # Starting address of ROM
    erase_weight: int        # Time it takes to perform an erase all
    crc_supported: int       # Is the function compute_crcs supported?

class Flash:
    """
    @brief Low-level control of flash programming algorithms.
    
    Instances of this class are bound to a flash memory region
    (@ref pyocd.core.memory_map.FlashRegion "FlashRegion") and support
    programming only within that region's address range. To program images that cross flash
    memory region boundaries, use the @ref pyocd.flash.loader.FlashLoader "FlashLoader" or
    @ref pyocd.flash.file_programmer.FileProgrammer "FileProgrammer" classes.
    
    Terminology:
    - sector: The size of an erasable block.
    - page: The size of a nominal programming block. Often flash can be programmed in much smaller
        increments (phrases). In that case, the page size determines the data buffers and the size
        passed to the ProgramPage() flash algo API. Pages must be the same size or smaller than
        sectors.
    - phrase: The minimum programming granularity, often from 1-16 bytes. For some flash
        technologies, the is no distinction between a phrase and a page.
    
    The `flash_algo` parameter of the constructor is a dictionary that defines all the details
    of the flash algorithm. The keys of this dictionary are as follows.
    - `load_address`: Memory address where the flash algo instructions will be loaded.
    - `instructions`: List of 32-bit words containing the position-independant code for the algo.
    - `pc_init`: Address of the `Init()` entry point. Optional.
    - `pc_eraseAll`: Address of the `EraseAll()` entry point. Optional.
    - `pc_erase_sector`: Address of the `EraseSector()` entry point.
    - `pc_program_page`: Address of the `ProgramPage()` entry point.
    - `pc_unInit`: Address of the `UnInit()` entry point. Optional.
    - `begin_data`: Base address of the page buffer. Used if `page_buffers` is not provided.
    - `page_buffers`: An optional list of base addresses for page buffers. The buffers must be at
        least as large as the region's page_size attribute. If at least 2 buffers are included in
        the list, then double buffered programming will be enabled.
    - `begin_stack`: Initial value of the stack pointer when calling any flash algo API.
    - `static_base`: Initial value of the R9 register for calling flash algo entry points, which
        determines where the position-independant data resides.
    - `analyzer_supported`: Whether the CRC32-based analyzer is supported.
    - `analyzer_address`: RAM base address where the analyzer code will be placed. There must be at
        least 0x600 free bytes after this address.
    
    All of the "pc_" entry point key values must have bit 0 set to indicate a Thumb function.
    """
    class Operation(Enum):
        """@brief Operations passed to init()."""
        ## Erase all or sector erase.
        ERASE = 1
        ## Program page or phrase.
        PROGRAM = 2
        ## Currently unused, but defined as part of the flash algorithm specification.
        VERIFY = 3

    ## Error value returned from wait_for_completion() on operation timeout.
    # 
    # The flash algo itself can never return a negative error code because the core register r0 is read
    # as unsigned.
    TIMEOUT_ERROR = -1

    def __init__(self, target, flash_algo):
        self.target = target
        self.flash_algo = flash_algo
        self.flash_algo_debug = False
        self._region = None
        self._did_prepare_target = False
        self._active_operation = None
        if flash_algo is not None:
            self.is_valid = True
            self.use_analyzer = flash_algo['analyzer_supported']
            self.end_flash_algo = flash_algo['load_address'] + len(flash_algo['instructions']) * 4
            self.begin_stack = flash_algo['begin_stack']
            self.begin_data = flash_algo['begin_data']
            self.static_base = flash_algo['static_base']
            self.min_program_length = flash_algo.get('min_program_length', 0)

            # Validate required APIs.
            assert self._is_api_valid('pc_erase_sector')
            assert self._is_api_valid('pc_program_page')

            # Check for double buffering support.
            if 'page_buffers' in flash_algo:
                self.page_buffers = flash_algo['page_buffers']
            else:
                self.page_buffers = [self.begin_data]

            self.double_buffer_supported = len(self.page_buffers) > 1

        else:
            self.is_valid = False
            self.use_analyzer = False
            self.end_flash_algo = None
            self.begin_stack = None
            self.begin_data = None
            self.static_base = None
            self.min_program_length = 0
            self.page_buffers = []
            self.double_buffer_supported = False
        
    def _is_api_valid(self, api_name):
        return (api_name in self.flash_algo) \
                and (self.flash_algo[api_name] >= self.flash_algo['load_address']) \
                and (self.flash_algo[api_name] < self.end_flash_algo)

    @property
    def minimum_program_length(self):
        assert self.region is not None
        return self.min_program_length or self.region.phrase_size

    @property
    def page_buffer_count(self):
        return len(self.page_buffers)
    
    @property
    def is_erase_all_supported(self):
        return self._is_api_valid('pc_eraseAll')

    @property
    def is_double_buffering_supported(self):
        return self.double_buffer_supported
    
    @property
    def region(self):
        return self._region
    
    @region.setter
    def region(self, flashRegion):
        assert flashRegion.is_flash
        self._region = flashRegion

    def init(self, operation, address=None, clock=0, reset=True):
        """!
        @brief Prepare the flash algorithm for performing operations.
        
        First, the target is prepared to execute flash algo operations, including loading the algo
        to target RAM. This step is skipped if the target is already prepared, i.e., init() has been
        called, but cleanup() not called yet.
        
        Next, the algo's Init() function is called with the provided parameters. If the algo does
        not have an Init() function, this step is skipped. Calling Init() is also skipped if the
        algo was previously inited for the same operation without an intervening uninit. If the
        algo is already inited for a different operation, uninit() is automatically called prior
        to intiting for the new operation.
        
        @exception FlashFailure
        """
        if address is None:
            address = self.get_flash_info().rom_start
        
        assert isinstance(operation, self.Operation)
        assert (self._did_prepare_target) or (not self._did_prepare_target and self._active_operation is None)
        
        self.target.halt()
        
        # Handle the algo already being inited.
        if self._active_operation is not None:
            # Uninit if the algo was left inited for a different operation.
            if self._active_operation != operation:
                self.uninit()
            # Don't need to reinit for the same operation.
            else:
                return

        # Setup target for running the flash algo.
        if not self._did_prepare_target:
            TRACE.debug("algo init and load to %#010x", self.flash_algo['load_address'])
            
            if reset:
                self.target.reset_and_halt(Target.ResetType.SW)
            self.prepare_target()

            # Load flash algo code into target RAM.
            self.target.write_memory_block32(self.flash_algo['load_address'], self.flash_algo['instructions'])

            self._did_prepare_target = True

        # update core register to execute the init subroutine
        if self._is_api_valid('pc_init'):
            TRACE.debug("algo call init(addr=%d, clock=%d, op=%d)", address, clock, operation.value)
            result = self._call_function_and_wait(self.flash_algo['pc_init'],
                                              r0=address, r1=clock, r2=operation.value, init=True,
                timeout=self.target.session.options.get('flash.timeout.init'))

            # check the return code
            TRACE.debug("init result = %d", result)
            if result == self.TIMEOUT_ERROR:
                raise FlashFailure('flash init timed out')
            elif result != 0:
                raise FlashFailure('flash init failure', result_code=result)
        
        self._active_operation = operation

    def cleanup(self):
        """! @brief Deinitialize the flash algo and restore the target.
        
        Before further operations are executed, the algo must be reinited. Unlike uninit(), this
        method marks the target and unprepared to execute flash algo functions. So on the next call
        to init(), the target will be prepared and the algo loaded into RAM.
        """
        self.uninit()
        self.restore_target()
        self._did_prepare_target = False

    def uninit(self):
        """! @brief Uninitialize the flash algo.
        
        Before further operations are executed, the algo must be reinited. The target is left in
        a state where algo does not have to be reloaded when init() is called.
        
        @exception FlashFailure
        """
        if self._active_operation is None:
            return
        
        if self._is_api_valid('pc_unInit'):
            TRACE.debug("call uninit(%d)", self._active_operation.value)

            # update core register to execute the uninit subroutine
            result = self._call_function_and_wait(self.flash_algo['pc_unInit'],
                                                    r0=self._active_operation.value,
                timeout=self.target.session.options.get('flash.timeout.init'))
            
            # check the return code
            TRACE.debug("uninit result = %d", result)
            if result == self.TIMEOUT_ERROR:
                raise FlashFailure('flash uninit timed out')
            elif result != 0:
                raise FlashFailure('flash uninit', result_code=result)
            
        self._active_operation = None

    def prepare_target(self):
        """! @brief Subclasses can override this method to perform special target configuration."""
        pass
    
    def restore_target(self):
        """! @brief Subclasses can override this method to undo any target configuration changes."""
        pass

    def compute_crcs(self, sectors):
        assert self.use_analyzer
        
        data = []

        # Load analyzer code into target RAM.
        self.target.write_memory_block32(self.flash_algo['analyzer_address'], _ANALYZER_CODE)

        # Convert address, size pairs into commands
        # for the crc computation algorithm to preform
        for addr, size in sectors:
            size_val = msb(size)
            addr_val = addr // size
            # Size must be a power of 2
            assert (1 << size_val) == size
            # Address must be a multiple of size
            assert (addr % size) == 0
            val = (size_val << 0) | (addr_val << 16)
            data.append(val)

        self.target.write_memory_block32(self.begin_data, data)

        # update core register to execute the subroutine
        TRACE.debug("call compute crc(%x, %x)", self.begin_data, len(data))
        self._call_function_and_wait(self.flash_algo['analyzer_address'], self.begin_data, len(data),
                timeout=self.target.session.options.get('flash.timeout.analyzer'))

        # Read back the CRCs for each section
        data = self.target.read_memory_block32(self.begin_data, len(data))
        return data

    def erase_all(self):
        """!
        @brief Erase all the flash.
        
        @exception FlashEraseFailure
        """
        assert self._active_operation == self.Operation.ERASE
        assert self.is_erase_all_supported

        # update core register to execute the erase_all subroutine
        TRACE.debug("call erase_all")
        result = self._call_function_and_wait(self.flash_algo['pc_eraseAll'],
                timeout=self.target.session.options.get('flash.timeout.erase_all'))

        # check the return code
        TRACE.debug("erase_all result = %d", result)
        if result == self.TIMEOUT_ERROR:
            raise FlashEraseFailure('flash erase all timed out')
        elif result != 0:
            raise FlashEraseFailure('flash erase all failure', result_code=result)

    def erase_sector(self, address):
        """!
        @brief Erase one sector.
        
        @exception FlashEraseFailure
        """
        assert self._active_operation == self.Operation.ERASE

        # update core register to execute the erase_sector subroutine
        TRACE.debug("call erase_sector(%x)", address)
        result = self._call_function_and_wait(self.flash_algo['pc_erase_sector'], address,
                timeout=self.target.session.options.get('flash.timeout.erase_sector'))

        # check the return code
        TRACE.debug("erase_sector result = %d", result)
        if result == self.TIMEOUT_ERROR:
            raise FlashEraseFailure('flash erase sector timed out')
        elif result != 0:
            raise FlashEraseFailure('flash erase sector failure', address=address, result_code=result)

    def program_page(self, address, bytes):
        """!
        @brief Flash one or more pages.
        
        @exception FlashProgramFailure
        """
        assert self._active_operation == self.Operation.PROGRAM

        # prevent security settings from locking the device
        bytes = self.override_security_bits(address, bytes)

        # first transfer in RAM
        self.target.write_memory_block8(self.begin_data, bytes)

        # update core register to execute the program_page subroutine
        TRACE.debug("call program_page(addr=%x, len=%x, data=%x)", address, len(bytes), self.begin_data)
        result = self._call_function_and_wait(self.flash_algo['pc_program_page'], address, len(bytes), self.begin_data,
                timeout=self.target.session.options.get('flash.timeout.program'))

        # check the return code
        TRACE.debug("program_page result = %d", result)
        if result == self.TIMEOUT_ERROR:
            raise FlashProgramFailure('flash program page timed out')
        elif result != 0:
            raise FlashProgramFailure('flash program page failure', address=address, result_code=result)

    def start_program_page_with_buffer(self, buffer_number, address):
        """!
        @brief Start flashing one or more pages.
        """
        assert self.region is not None
        assert buffer_number < len(self.page_buffers), "Invalid buffer number"
        assert self._active_operation == self.Operation.PROGRAM

        # update core register to execute the program_page subroutine
        TRACE.debug("start_program_page_with_buffer(addr=%x, len=%x, data=%x)", address, self.region.page_size,
                self.page_buffers[buffer_number])
        self._call_function(self.flash_algo['pc_program_page'], address, self.region.page_size, self.page_buffers[buffer_number])

    def load_page_buffer(self, buffer_number, address, bytes):
        """!
        @brief Load data to a numbered page buffer.
        
        This method is used in conjunction with start_program_page_with_buffer() to implement
        double buffered programming.
        """
        assert buffer_number < len(self.page_buffers), "Invalid buffer number"

        # prevent security settings from locking the device
        bytes = self.override_security_bits(address, bytes)

        # transfer the buffer to device RAM
        self.target.write_memory_block8(self.page_buffers[buffer_number], bytes)

    def program_phrase(self, address, bytes):
        """!
        @brief Flash a portion of a page.
        
        @exception FlashFailure The address or data length is not aligned to the minimum
            programming length specified in the flash algorithm.
        @exception FlashProgramFailure
        """
        assert self._active_operation == self.Operation.PROGRAM

        # Get min programming length.
        min_len = self.minimum_program_length

        # Require write address and length to be aligned to min write size.
        if address % min_len:
            raise FlashFailure("flash program phrase failure: unaligned start address", address=address)
        if len(bytes) % min_len:
            raise FlashFailure("flash program phrase failure: phrase length is unaligned or too small", address=address)

        # prevent security settings from locking the device
        bytes = self.override_security_bits(address, bytes)

        # first transfer in RAM
        self.target.write_memory_block8(self.begin_data, bytes)

        # update core register to execute the program_page subroutine
        TRACE.debug("call program_phrase(addr=%x, len=%x, data=%x)", address, len(bytes), self.begin_data)
        result = self._call_function_and_wait(self.flash_algo['pc_program_page'], address, len(bytes), self.begin_data,
                timeout=self.target.session.options.get('flash.timeout.program'))

        # check the return code
        if result == self.TIMEOUT_ERROR:
            raise FlashProgramFailure('flash program phrase timed out')
        elif result != 0:
            raise FlashProgramFailure('flash program phrase failure', address=address, result_code=result)

    def get_sector_info(self, addr):
        """!
        @brief Get info about the sector that contains this address.
        """
        assert self.region is not None
        if not self.region.contains_address(addr):
            return None

        info = SectorInfo(
                erase_weight=self.region.erase_sector_weight,
                size=self.region.sector_size,
                base_addr=align_down(addr, self.region.sector_size),
                )
        return info

    def get_page_info(self, addr):
        """!
        @brief Get info about the page that contains this address.
        """
        assert self.region is not None
        if not self.region.contains_address(addr):
            return None

        info = PageInfo(
                program_weight=self.region.program_page_weight,
                size=self.region.page_size,
                base_addr=align_down(addr, self.region.page_size)
                )
        return info

    def get_flash_info(self):
        """!
        @brief Get info about the flash.

        Override this method to return different values.
        """
        assert self.region is not None

        info = FlashInfo(
                rom_start=self.region.start,
                erase_weight=self.region.erase_all_weight,
                crc_supported=self.use_analyzer,
                )
        return info

    def get_flash_builder(self):
        return FlashBuilder(self)

    def flash_block(self, addr, data, smart_flash=True, chip_erase=None, progress_cb=None, fast_verify=False):
        """!
        @brief Flash a block of data.

        @note Deprecated. Will be removed in v1.0.
        """
        assert self.region is not None
        assert self.region.contains_range(start=addr, length=len(data))
        
        fb = FlashBuilder(self)
        fb.add_data(addr, data)
        info = fb.program(chip_erase, progress_cb, smart_flash, fast_verify)
        return info

    def _call_function(self, pc, r0=None, r1=None, r2=None, r3=None, init=False):
        reg_list = []
        data_list = []

        if self.flash_algo_debug:
            self._flash_algo_debug_setup()

        reg_list.append('pc')
        data_list.append(pc)
        if r0 is not None:
            reg_list.append('r0')
            data_list.append(r0)
        if r1 is not None:
            reg_list.append('r1')
            data_list.append(r1)
        if r2 is not None:
            reg_list.append('r2')
            data_list.append(r2)
        if r3 is not None:
            reg_list.append('r3')
            data_list.append(r3)
        if init:
            reg_list.append('r9')
            data_list.append(self.static_base)
        if init:
            reg_list.append('sp')
            data_list.append(self.begin_stack)
        reg_list.append('lr')
        data_list.append(self.flash_algo['load_address'] + 1)
        self.target.write_core_registers_raw(reg_list, data_list)

        # resume target
        self.target.resume()

    def _flash_algo_debug_setup(self):
        # Save vector catch state for use in wait_for_completion()
        self._saved_vector_catch = self.target.get_vector_catch()
        self.target.set_vector_catch(Target.VectorCatch.ALL)

    def _flash_algo_debug_check(self):
        regs = self.target.read_core_registers_raw(list(range(19)) + [20])
        LOG.debug("Registers after flash algo: [%s]", " ".join("%08x" % r for r in regs))

        expected_fp = self.flash_algo['static_base']
        expected_sp = self.flash_algo['begin_stack']
        expected_pc = self.flash_algo['load_address']
        final_ipsr = self.target.read_core_register('ipsr')
        final_fp = self.target.read_core_register('r9')
        final_sp = self.target.read_core_register('sp')
        final_pc = self.target.read_core_register('pc')
        #TODO - uncomment if Read/write and zero init sections can be moved into a separate flash algo section
        #expected_flash_algo = self.flash_algo['instructions']
        #if self.use_analyzer:
        #    expected_analyzer = analyzer
        #final_flash_algo = self.target.read_memory_block32(self.flash_algo['load_address'], len(self.flash_algo['instructions']))
        #if self.use_analyzer:
        #    final_analyzer = self.target.read_memory_block32(self.flash_algo['analyzer_address'], len(analyzer))

        error = False
        if final_ipsr != 0:
            LOG.error("IPSR should be 0 but is 0x%x", final_ipsr)
            error = True
        if final_fp != expected_fp:
            # Frame pointer should not change
            LOG.error("Frame pointer should be 0x%x but is 0x%x" % (expected_fp, final_fp))
            error = True
        if final_sp != expected_sp:
            # Stack pointer should return to original value after function call
            LOG.error("Stack pointer should be 0x%x but is 0x%x" % (expected_sp, final_sp))
            error = True
        if final_pc != expected_pc:
            # PC should be pointing to breakpoint address
            LOG.error("PC should be 0x%x but is 0x%x" % (expected_pc, final_pc))
            error = True
        #TODO - uncomment if Read/write and zero init sections can be moved into a separate flash algo section
        #if not _same(expected_flash_algo, final_flash_algo):
        #    LOG.error("Flash algorithm overwritten!")
        #    error = True
        #if self.use_analyzer and not _same(expected_analyzer, final_analyzer):
        #    LOG.error("Analyzer overwritten!")
        #    error = True
        assert error == False
        self.target.set_vector_catch(self._saved_vector_catch)

    def wait_for_completion(self, timeout=None):
        """!
        @brief Wait until the breakpoint is hit.
        """
        with Timeout(timeout) as time_out:
            while time_out.check():
                if self.target.get_state() != Target.State.RUNNING:
                    break
            else:
                # Operation timed out. 
                self.target.halt()
                return self.TIMEOUT_ERROR

        if self.flash_algo_debug:
            self._flash_algo_debug_check()

        return self.target.read_core_register('r0')

    def _call_function_and_wait(self, pc, r0=None, r1=None, r2=None, r3=None, init=False, timeout=None):
        self._call_function(pc, r0, r1, r2, r3, init)
        return self.wait_for_completion(timeout=timeout)

    def set_flash_algo_debug(self, enable):
        """!
        @brief Turn on extra flash algorithm checking

        When set this may slow down flash algo performance.
        """
        self.flash_algo_debug = enable

    def override_security_bits(self, address, data):
        return data



# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2019 Monadnock Systems Ltd.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep

from ...core import exceptions
from ...coresight.coresight_target import CoreSightTarget
from ...utility.timeout import Timeout

AHB_AP_NUM = 0x0
CTRL_AP_NUM = 0x1

CTRL_AP_RESET = 0x000
CTRL_AP_ERASEALL = 0x004
CTRL_AP_ERASEALLSTATUS = 0x008
CTRL_AP_APPROTECTSTATUS = 0x00C
CTRL_AP_IDR = 0x0FC

CTRL_AP_ERASEALLSTATUS_READY = 0x0
CTRL_AP_ERASEALLSTATUS_BUSY = 0x1

CTRL_AP_APPROTECTSTATUS_ENABLED = 0x0
CTRL_AP_APPROTECTSTATUS_DISABLED = 0x1

CTRL_AP_RESET_NORESET = 0x0
CTRL_AP_RESET_RESET = 0x1

CTRL_AP_ERASEALL_NOOPERATION = 0x0
CTRL_AP_ERASEALL_ERASE = 0x1

CTRL_IDR_EXPECTED = 0x2880000
CTRL_IDR_VERSION_MASK = 0xf0000000
CTRL_IDR_VERSION_SHIFT = 28

MASS_ERASE_TIMEOUT = 15.0

LOG = logging.getLogger(__name__)


class NRF52(CoreSightTarget):

    VENDOR = "Nordic Semiconductor"

    def __init__(self, session, memory_map=None):
        super(NRF52, self).__init__(session, memory_map)
        self.ctrl_ap = None

    def create_init_sequence(self):
        seq = super(NRF52, self).create_init_sequence()

        # Must check whether security is enabled, and potentially auto-unlock, before
        # any init tasks that require system bus access.
        seq.wrap_task('discovery',
            lambda seq: seq.insert_before('find_components',
                              ('check_ctrl_ap_idr', self.check_ctrl_ap_idr),
                              ('check_flash_security', self.check_flash_security),
                          )
            )

        return seq

    def check_ctrl_ap_idr(self):
        self.ctrl_ap = self.dp.aps[CTRL_AP_NUM]

        # Check CTRL-AP ID.
        if (self.ctrl_ap.idr & ~CTRL_IDR_VERSION_MASK) != CTRL_IDR_EXPECTED:
            LOG.error("%s: bad CTRL-AP IDR (is 0x%08x)", self.part_number, self.ctrl_ap.idr)

        ctrl_ap_version = (self.ctrl_ap.idr & CTRL_IDR_VERSION_MASK) >> CTRL_IDR_VERSION_SHIFT
        LOG.debug("CTRL-AP version %d", ctrl_ap_version)

    def check_flash_security(self):
        """! @brief Check security and unlock device.

        This init task determines whether the device is locked (APPROTECT enabled). If it is,
        and if auto unlock is enabled, then perform a mass erase to unlock the device.

        This init task runs *before* cores are created.
        """

        if self.is_locked():
            if self.session.options.get('auto_unlock'):
                LOG.warning("%s APPROTECT enabled: will try to unlock via mass erase", self.part_number)

                # Do the mass erase.
                if not self.mass_erase():
                    LOG.error("%s: mass erase failed", self.part_number)
                    raise exceptions.TargetError("unable to unlock device")
                # Cached badness from create_ap run during AP lockout prevents create_cores from
                # succeeding.
                self.dp.create_1_ap(AHB_AP_NUM)
            else:
                LOG.warning("%s APPROTECT enabled: not automatically unlocking", self.part_number)
        else:
            LOG.info("%s not in secure state", self.part_number)

    def is_locked(self):
        status = self.ctrl_ap.read_reg(CTRL_AP_APPROTECTSTATUS)
        return status == CTRL_AP_APPROTECTSTATUS_ENABLED

    def mass_erase(self):
        # See Nordic Whitepaper nWP-027 for magic numbers and order of operations from the vendor
        self.ctrl_ap.write_reg(CTRL_AP_ERASEALL, CTRL_AP_ERASEALL_ERASE)
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                status = self.ctrl_ap.read_reg(CTRL_AP_ERASEALLSTATUS)
                if status == CTRL_AP_ERASEALLSTATUS_READY:
                    break
                sleep(0.1)
            else:
                # Timed out
                LOG.error("Mass erase timeout waiting for ERASEALLSTATUS")
                return False
        self.ctrl_ap.write_reg(CTRL_AP_RESET, CTRL_AP_RESET_RESET)
        self.ctrl_ap.write_reg(CTRL_AP_RESET, CTRL_AP_RESET_NORESET)
        self.ctrl_ap.write_reg(CTRL_AP_ERASEALL, CTRL_AP_ERASEALL_NOOPERATION)
        return True



# pyOCD debugger
# Copyright (c) 2017-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from .cmsis_pack import (CmsisPack, MalformedCmsisPackError)
from ..family import FAMILIES
from .. import TARGET
from ...coresight.coresight_target import CoreSightTarget
from ...debug.svd.loader import SVDFile

try:
    import cmsis_pack_manager
    CPM_AVAILABLE = True
except ImportError:
    CPM_AVAILABLE = False

LOG = logging.getLogger(__name__)

class ManagedPacks(object):
    """! @brief Namespace for managed CMSIS-Pack utilities.
    
    By managed, we mean managed by the cmsis-pack-manager package. All the methods on this class
    apply only to those packs managed by cmsis-pack-manager, not any targets from packs specified
    by the user.
    """

    @staticmethod
    def get_installed_packs(cache=None):
        """! @brief Return a list containing CmsisPackRef objects for all installed packs."""
        if not CPM_AVAILABLE:
            return []
        if cache is None:
            cache = cmsis_pack_manager.Cache(True, True)
        results = []
        # packs_for_devices() returns only unique packs.
        for pack in cache.packs_for_devices(cache.index.values()):
            # Generate full path to the .pack file.
            pack_path = os.path.join(cache.data_path, pack.get_pack_name())
            
            # If the .pack file exists, the pack is installed.
            if os.path.isfile(pack_path):
                results.append(pack)
        return results

    @staticmethod
    def get_installed_targets(cache=None):
        """! @brief Return a list of CmsisPackDevice objects for installed pack targets."""
        if not CPM_AVAILABLE:
            return []
        if cache is None:
            cache = cmsis_pack_manager.Cache(True, True)
        results = []
        for pack in ManagedPacks.get_installed_packs(cache=cache):
            pack_path = os.path.join(cache.data_path, pack.get_pack_name())
            pack = CmsisPack(pack_path)
            results += list(pack.devices)
        return sorted(results, key=lambda dev:dev.part_number)

    @staticmethod
    def populate_target(device_name):
        """! @brief Add targets from cmsis-pack-manager matching the given name.

        Targets are added to the `#TARGET` list. A case-insensitive comparison against the
        device part number is used to find the target to populate. If multiple packs are installed
        that provide the same part numbers, all matching targets will be populated.
        """
        device_name = device_name.lower()
        targets = ManagedPacks.get_installed_targets()
        for dev in targets:
            if device_name == dev.part_number.lower():
                PackTargets.populate_device(dev)

class _PackTargetMethods(object):
    """! @brief Container for methods added to the dynamically generated pack target subclass."""

    @staticmethod
    def _pack_target__init__(self, session):
        """! @brief Constructor for dynamically created target class."""
        super(self.__class__, self).__init__(session, self._pack_device.memory_map)

        self.vendor = self._pack_device.vendor
        self.part_families = self._pack_device.families
        self.part_number = self._pack_device.part_number

        self._svd_location = SVDFile(filename=self._pack_device.svd)

    @staticmethod
    def _pack_target_create_init_sequence(self):
        """! @brief Creates an init task to set the default reset type."""
        seq = super(self.__class__, self).create_init_sequence()
        seq.wrap_task('discovery',
            lambda seq: seq.insert_after('create_cores',
                            ('set_default_reset_type', self.set_default_reset_type)
                            )
            )
        return seq

    @staticmethod
    def _pack_target_set_default_reset_type(self):
        """! @brief Set's the first core's default reset type to the one specified in the pack."""
        if 0 in self.cores:
            self.cores[0].default_reset_type = self._pack_device.default_reset_type

class PackTargets(object):
    """! @brief Namespace for CMSIS-Pack target generation utilities. """

    @staticmethod
    def _find_family_class(dev):
        """! @brief Search the families list for matching entry."""
        for familyInfo in FAMILIES:
            # Skip if wrong vendor.
            if dev.vendor != familyInfo.vendor:
                continue

            # Scan each level of family plus part number, from specific to generic.
            for compare_name in reversed(dev.families + [dev.part_number]):
                # Require the regex to match the entire family name.
                match = familyInfo.matches.match(compare_name)
                if match and match.span() == (0, len(compare_name)):
                    return familyInfo.klass

        # Didn't match, so return default target superclass.
        return CoreSightTarget

    @staticmethod
    def _generate_pack_target_class(dev):
        """! @brief Generates a new target class from a CmsisPackDevice.

        @param dev A CmsisPackDevice object.
        @return A new subclass of either CoreSightTarget or one of the family classes.
        """
        try:
            # Look up the target family superclass.
            superklass = PackTargets._find_family_class(dev)

            # Replace spaces and dashes with underscores on the new target subclass name.
            subclassName = dev.part_number.replace(' ', '_').replace('-', '_')

            # Create a new subclass for this target.
            targetClass = type(subclassName, (superklass,), {
                        "_pack_device": dev,
                        "__init__": _PackTargetMethods._pack_target__init__,
                        "create_init_sequence": _PackTargetMethods._pack_target_create_init_sequence,
                        "set_default_reset_type": _PackTargetMethods._pack_target_set_default_reset_type,
                    })
            return targetClass
        except (MalformedCmsisPackError, FileNotFoundError) as err:
            LOG.warning(err)
            return None

    @staticmethod
    def populate_device(dev):
        """! @brief Generates and populates the target defined by a CmsisPackDevice.

        The new target class is added to the `#TARGET` list.

        @param dev A CmsisPackDevice object.
        """
        try:
            tgt = PackTargets._generate_pack_target_class(dev)
            if tgt is None:
                return
            part = dev.part_number.lower().replace("-", "_")

            # Make sure there isn't a duplicate target name.
            if part not in TARGET:
                TARGET[part] = tgt
            else:
                LOG.debug("did not populate %s because there is already a %s target installed", dev.part_number, part)
        except (MalformedCmsisPackError, FileNotFoundError) as err:
            LOG.warning(err)

    @staticmethod
    def populate_targets_from_pack(pack_list):
        """! @brief Adds targets defined in the provided CMSIS-Pack.

        Targets are added to the `#TARGET` list.

        @param pack_list Sequence of strings that are paths to .pack files, file objects,
            ZipFile instances, or CmsisPack instance. May also be a single object of one of
            the accepted types.
        """
        if not isinstance(pack_list, (list, tuple)):
            pack_list = [pack_list]
        for pack_or_path in pack_list:
            if isinstance(pack_or_path, CmsisPack):
                pack = pack_or_path
            else:
                pack = CmsisPack(pack_or_path)
            for dev in pack.devices:
                PackTargets.populate_device(dev)



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .target import Target
from ..flash.eraser import FlashEraser
from ..debug.cache import CachingDebugContext
from ..debug.elf.elf import ELFBinaryFile
from ..debug.elf.elf_reader import ElfReaderContext
from ..utility.graph import GraphNode
from ..utility.sequencer import CallSequence

LOG = logging.getLogger(__name__)

class SoCTarget(Target, GraphNode):
    """! @brief Represents a microcontroller system-on-chip.
    
    An instance of this class is the root of the chip-level object graph. It has child
    nodes for the DP and all cores. As a concrete subclass of Target, it provides methods
    to control the device, access memory, adjust breakpoints, and so on.
    
    For single core devices, the SoCTarget has mostly equivalent functionality to
    the Target object for the core. Multicore devices work differently. This class tracks
    a "selected core", to which all actions are directed. The selected core can be changed
    at any time. You may also directly access specific cores and perform operations on them.
    """
    
    VENDOR = "Generic"

    def __init__(self, session, memory_map=None):
        Target.__init__(self, session, memory_map)
        GraphNode.__init__(self)
        self.vendor = self.VENDOR
        self.part_families = getattr(self, 'PART_FAMILIES', [])
        self.part_number = getattr(self, 'PART_NUMBER', self.__class__.__name__)
        self._cores = {}
        self._selected_core = None
        self._new_core_num = 0
        self._elf = None

    @property
    def cores(self):
        return self._cores

    @property
    def selected_core(self):
        if self._selected_core is None:
            return None
        return self.cores[self._selected_core]
    
    @selected_core.setter
    def selected_core(self, core_number):
        if core_number not in self.cores:
            raise ValueError("invalid core number %d" % core_number)
        LOG.debug("selected core #%d" % core_number)
        self._selected_core = core_number

    @property
    def elf(self):
        return self._elf

    @elf.setter
    def elf(self, filename):
        if filename is None:
            self._elf = None
        else:
            self._elf = ELFBinaryFile(filename, self.memory_map)
            for core_number in range(len(self.cores)):
                self.cores[core_number].elf = self._elf
                if self.session.options['cache.read_code_from_elf']:
                    self.cores[core_number].set_target_context(
                            ElfReaderContext(self.cores[core_number].get_target_context(), self._elf))
    
    @property
    def supported_security_states(self):
        return self.selected_core.supported_security_states
    
    @property
    def core_registers(self):
        return self.selected_core.core_registers

    def add_core(self, core):
        core.delegate = self.delegate
        core.set_target_context(CachingDebugContext(core))
        self.cores[core.core_number] = core
        self.add_child(core)
        
        if self._selected_core is None:
            self._selected_core = core.core_number

    def create_init_sequence(self):
        # Return an empty call sequence. The subclass must override this.
        return CallSequence()
    
    def init(self):
        # If we don't have a delegate installed yet but there is a session delegate, use it.
        if (self.delegate is None) and (self.session.delegate is not None):
            self.delegate = self.session.delegate
        
        # Create and execute the init sequence.
        seq = self.create_init_sequence()
        self.call_delegate('will_init_target', target=self, init_sequence=seq)
        seq.invoke()
        self.call_delegate('did_init_target', target=self)
    
    def post_connect_hook(self):
        """! @brief Hook function called after post_connect init task.
        
        This hook lets the target subclass configure the target as necessary.
        """
        pass

    def disconnect(self, resume=True):
        self.session.notify(Target.Event.PRE_DISCONNECT, self)
        self.call_delegate('will_disconnect', target=self, resume=resume)
        for core in self.cores.values():
            core.disconnect(resume)
        self.call_delegate('did_disconnect', target=self, resume=resume)

    @property
    def run_token(self):
        return self.selected_core.run_token

    def halt(self):
        return self.selected_core.halt()

    def step(self, disable_interrupts=True, start=0, end=0, hook_cb=None):
        return self.selected_core.step(disable_interrupts, start, end, hook_cb)

    def resume(self):
        return self.selected_core.resume()

    def mass_erase(self):
        if not self.call_delegate('mass_erase', target=self):
            # The default mass erase implementation is to simply perform a chip erase.
            FlashEraser(self.session, FlashEraser.Mode.CHIP).erase()
        return True

    def write_memory(self, addr, value, transfer_size=32):
        return self.selected_core.write_memory(addr, value, transfer_size)

    def read_memory(self, addr, transfer_size=32, now=True):
        return self.selected_core.read_memory(addr, transfer_size, now)

    def write_memory_block8(self, addr, value):
        return self.selected_core.write_memory_block8(addr, value)

    def write_memory_block32(self, addr, data):
        return self.selected_core.write_memory_block32(addr, data)

    def read_memory_block8(self, addr, size):
        return self.selected_core.read_memory_block8(addr, size)

    def read_memory_block32(self, addr, size):
        return self.selected_core.read_memory_block32(addr, size)

    def read_core_register(self, id):
        return self.selected_core.read_core_register(id)

    def write_core_register(self, id, data):
        return self.selected_core.write_core_register(id, data)

    def read_core_register_raw(self, reg):
        return self.selected_core.read_core_register_raw(reg)

    def read_core_registers_raw(self, reg_list):
        return self.selected_core.read_core_registers_raw(reg_list)

    def write_core_register_raw(self, reg, data):
        self.selected_core.write_core_register_raw(reg, data)

    def write_core_registers_raw(self, reg_list, data_list):
        self.selected_core.write_core_registers_raw(reg_list, data_list)

    def find_breakpoint(self, addr):
        return self.selected_core.find_breakpoint(addr)

    def set_breakpoint(self, addr, type=Target.BreakpointType.AUTO):
        return self.selected_core.set_breakpoint(addr, type)

    def get_breakpoint_type(self, addr):
        return self.selected_core.get_breakpoint_type(addr)

    def remove_breakpoint(self, addr):
        return self.selected_core.remove_breakpoint(addr)

    def set_watchpoint(self, addr, size, type):
        return self.selected_core.set_watchpoint(addr, size, type)

    def remove_watchpoint(self, addr, size, type):
        return self.selected_core.remove_watchpoint(addr, size, type)

    def reset(self, reset_type=None):
        # Use the probe to reset to perform a hardware reset if there is not a core.
        if self.selected_core is None:
            # Use the probe to reset. (We can't use the DP here because that's a class layering violation;
            # the DP is only created by the CoreSightTarget subclass.)
            self.session.probe.reset()
            return
        self.selected_core.reset(reset_type)

    def reset_and_halt(self, reset_type=None):
        return self.selected_core.reset_and_halt(reset_type)

    def get_state(self):
        return self.selected_core.get_state()
        
    def get_security_state(self):
        return self.selected_core.get_security_state()

    def get_halt_reason(self):
        return self.selected_core.get_halt_reason()

    def set_vector_catch(self, enableMask):
        return self.selected_core.set_vector_catch(enableMask)

    def get_vector_catch(self):
        return self.selected_core.get_vector_catch()

    def get_target_context(self, core=None):
        if core is None:
            core = self._selected_core
        return self.cores[core].get_target_context()
    
    def trace_start(self):
        self.call_delegate('trace_start', target=self, mode=0)
    
    def trace_stop(self):
        self.call_delegate('trace_stop', target=self, mode=0)
    
        



# pyOCD debugger
# Copyright (c) 2021 Huada Semiconductor Corporation
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile


#DEBUG_ACTIVE
DEBUG_ACTIVE = 0x40002038
DEBUG_ACTIVE_VAL = 0x00000FFF

FLASH_ALGO = { 
    'load_address' : 0x20000000,
    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x4770ba40, 0x4770bac0, 0x6a014809, 0xd4fc06c9, 0x62c14908, 0x62c14908, 0x22036a01, 0x62014311,
    0x60092100, 0x06c96a01, 0x2000d4fc, 0x00004770, 0x40020000, 0x00005a5a, 0x0000a5a5, 0x6a0a4909,
    0xd4fc06d2, 0x62ca4a08, 0x62ca4a08, 0x08926a0a, 0x1c920092, 0x2200620a, 0x6a086002, 0xd4fc06c0,
    0x47702000, 0x40020000, 0x00005a5a, 0x0000a5a5, 0x2000b438, 0x70084669, 0x6801484a, 0x0f4a0609,
    0x700a4669, 0x89094948, 0x054968c2, 0x0ad20d49, 0x430a02d2, 0x4a4560c2, 0x49456082, 0x68036081,
    0x022d2507, 0x600343ab, 0x60816082, 0x24016803, 0x60034323, 0x051b68c3, 0x466bd5fc, 0x2b00781b,
    0x6082d025, 0x68036081, 0x43a324e0, 0x60826003, 0x466b6081, 0x2b01781b, 0x2b02d006, 0x2b03d009,
    0x2b04d00c, 0xe00ed113, 0x24026803, 0x600343a3, 0x6803e00d, 0x43a32404, 0xe0086003, 0x24086803,
    0x600343a3, 0x6803e003, 0x43a32410, 0x60826003, 0x68036081, 0x600343ab, 0x60816082, 0x24036803,
    0x43a302e4, 0x6a036003, 0x43230524, 0x48216203, 0x62c162c2, 0x240c6a03, 0x620343a3, 0x62c162c2,
    0x60032320, 0x62c162c2, 0x60432317, 0x62c162c2, 0x6083231b, 0x62c162c2, 0x60c34b17, 0x62c162c2,
    0x61034b16, 0x62c162c2, 0x61432318, 0x62c162c2, 0x618323f0, 0x62c162c2, 0x00db237d, 0x62c261c3,
    0x230062c1, 0x630343db, 0x62c162c2, 0x62c26343, 0x4c0b62c1, 0x62c26023, 0x606362c1, 0x2000bc38,
    0x00004770, 0x40002000, 0x00100c00, 0x00005a5a, 0x0000a5a5, 0x40020000, 0x00004650, 0x000222e0,
    0x40020040, 0x4b0db430, 0x25004c0b, 0x4c0c62dc, 0x6a1c62dc, 0x00a408a4, 0x621c1c64, 0xd9072900,
    0x55445d54, 0x06e46a1c, 0x1c6dd4fc, 0xd3f7428d, 0x2000bc30, 0x00004770, 0x00005a5a, 0x40020000,
    0x0000a5a5, 0x480bb410, 0x62c14909, 0x62c24a0a, 0x63032300, 0x62c262c1, 0x62c16343, 0x4c0762c2,
    0x62c16023, 0x606362c2, 0x4618bc10, 0x00004770, 0x00005a5a, 0x40020000, 0x0000a5a5, 0x40020040,
    0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000091,
    'pc_unInit': 0x20000225,
    'pc_program_page': 0x200001e5,
    'pc_erase_sector': 0x2000005d,
    'pc_eraseAll': 0x20000029,

    'static_base' : 0x20000000 + 0x00000020 + 0x00000240,
    'begin_stack' : 0x20000500,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x200,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001200],   # Enable double buffering
    'min_program_length' : 0x200,
  }


class HC32L196(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x40000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x8000)
        )

    def __init__(self, session):
        super(HC32L196, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L19x.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32L190(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x40000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x8000)
        )

    def __init__(self, session):
        super(HC32L190, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L19x.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32F190(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x40000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x8000)
        )

    def __init__(self, session):
        super(HC32F190, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L19x.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32F196(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x40000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x8000)
        )

    def __init__(self, session):
        super(HC32F196, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L19x.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyocd.utility.strings import (
    uniquify_name,
    )

class TestUniquifyName:
    def test_empty_with_no_others(self):
        assert uniquify_name('', []) == ''
    
    def test_empty_with_others(self):
        assert uniquify_name('', ['bar', 'buz']) == ''
    
    def test_empty_with_another_empty(self):
        assert uniquify_name('', ['bar', 'buz', '']) == '_1'
    
    def test_no_others(self):
        assert uniquify_name('foo', []) == 'foo'
    
    def test_already_unique(self):
        assert uniquify_name('foo', ['bar']) == 'foo'
    
    def test_no_trailing_int(self):
        assert uniquify_name('foo', ['foo']) == 'foo_1'
    
    def test_1_trailing_int(self):
        assert uniquify_name('foo1', ['foo1']) == 'foo2'
    
    def test_multiple_trailing_ints(self):
        assert uniquify_name('foo1', ['foo1', 'foo2']) == 'foo3'
    
    def test_name_has_int(self):
        assert uniquify_name('foo2', ['foo2', 'bar', 'foo3']) == 'foo4'
    
    def test_multiple_ints_in_name(self):
        assert uniquify_name('baz 2 monkey-3', ['fun', 'baz 2 monkey-3']) == 'baz 2 monkey-4'




# pyOCD debugger
# Copyright (c) 2020-2021 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import json
import socket
from socketserver import (ThreadingTCPServer, StreamRequestHandler)

from .shared_probe_proxy import SharedDebugProbeProxy
from ..core import exceptions
from .debug_probe import DebugProbe
from ..coresight.ap import (APVersion, APv1Address, APv2Address)

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class DebugProbeServer(threading.Thread):
    """! @brief Shares a debug probe over a TCP server.
    
    When the start() method is called, a new daemon thread is created to run the server. The server
    can be terminated by calling the stop() method, which will also kill the server thread.
    """
    
    def __init__(self, session, probe, port=None, serve_local_only=None):
        """! @brief Constructor.
        
        @param self The object.
        @param session A @ref pyocd.core.session.Session "Session" object. Does not need to have a
            probe assigned to it.
        @param probe Either the @ref pyocd.probe.debug_probe.DebugProbe "DebugProbe" object to serve
            or a @ref pyocd.probe.shared_probe_proxy.SharedDebugProbeProxy "SharedDebugProbeProxy".
            Doesn't have to be associated with a session, and should not be opened already. If not
            already an instance of
            @ref pyocd.probe.shared_probe_proxy.SharedDebugProbeProxy "SharedDebugProbeProxy"
            then a new proxy is created to allow the probe to be shared by multiple connections.
        @param port The TCP port number. Defaults to the 'probeserver.port' option if not provided.
        @param serve_local_only Optional Boolean. Whether to restrict the server to be accessible only from
            localhost. If not specified (set to None), then the 'serve_local_only' session option is used.
        """
        super(DebugProbeServer, self).__init__()
        
        # Configure the server thread.
        self.name = "debug probe %s server" % probe.unique_id
        self.daemon = True
        
        # Init instance variables.
        self._session = session
        self._probe = probe
        self._is_running = False
        
        # Make sure we have a shared proxy for the probe.
        if isinstance(probe, SharedDebugProbeProxy):
            self._proxy = probe
        else:
            self._proxy = SharedDebugProbeProxy(probe)
        
        # Get the port from options if not specified.
        if port is None:
            self._port = session.options.get('probeserver.port')
        else:
            self._port = port
        
        # Default to the serve_local_only session option.
        if serve_local_only is None:
            serve_local_only = session.options.get('serve_local_only')
        
        host = 'localhost' if serve_local_only else ''
        address = (host, self._port)
        
        # Create the server and bind to the address, but don't start running yet.
        self._server = TCPProbeServer(address, session, self._proxy)
        self._server.server_bind()
        
    def start(self):
        """! @brief Start the server thread and begin listening."""
        self._server.server_activate()
        super(DebugProbeServer, self).start()
    
    def stop(self):
        """! @brief Shut down the server.
        
        Any open connections will be forcibly closed. This function does not return until the
        server thread has exited.
        """
        self._server.shutdown()
        self.join()
    
    @property
    def is_running(self):
        """! @brief Whether the server thread is running."""
        return self._is_running
    
    @property
    def port(self):
        """! @brief The server's port.
        
        If port 0 was specified in the constructor, then, after start() is called, this will reflect the actual port
        on which the server is listening.
        """
        return self._port
    
    def run(self):
        """! @brief The server thread implementation."""
        self._is_running = True
        
        # Read back the actual port if 0 was specified.
        if self._port == 0:
            self._port = self._server.socket.getsockname()[1]
        
        LOG.info("Serving debug probe %s (%s) on port %i",
                self._probe.description, self._probe.unique_id, self._port)
        self._server.serve_forever()
        self._is_running = False

class TCPProbeServer(ThreadingTCPServer):
    """! @brief TCP server subclass that carries the session and probe being served."""
    
    # Change the default SO_REUSEADDR setting.
    allow_reuse_address = True
    
    def __init__(self, server_address, session, probe):
        self._session = session
        self._probe = probe
        ThreadingTCPServer.__init__(self, server_address, DebugProbeRequestHandler,
            bind_and_activate=False)
    
    @property
    def session(self):
        return self._session
    
    @property
    def probe(self):
        return self._probe
    
    def handle_error(self, request, client_address):
        LOG.error("Error while handling client request (client address %s):", client_address,
            exc_info=self._session.log_tracebacks)

class DebugProbeRequestHandler(StreamRequestHandler):
    """!
    @brief Probe server request handler.
    
    This class implements the server side for the remote probe protocol.

    request:
    ````
    {
      "id": <int>,
      "request": <str>,
      ["arguments": <list>]
    }
    ````

    response:
    ````
    {
      "id": <int>,
      "status": <int>,
      ["error": <str>,]
      ["response": <value>]
    }
    ````
    """
    
    ## Current version of the remote probe protocol.
    PROTOCOL_VERSION = 1
    
    class StatusCode:
        """! @brief Constants for errors reported from the server."""
        GENERAL_ERROR = 1
        PROBE_DISCONNECTED = 2
        PROBE_ERROR = 3
        TRANSFER_ERROR = 10
        TRANSFER_TIMEOUT = 11
        TRANSFER_FAULT = 12

    def setup(self):
        # Do a DNS lookup on the client.
        try:
            info = socket.gethostbyaddr(self.client_address[0])
            self._client_domain = info[0]
        except socket.herror:
            self._client_domain = self.client_address[0]
        
        LOG.info("Remote probe client connected (%s from port %i)", self._client_domain, self.client_address[1])
        
        # Get the session and probe we're serving from the server.
        self._session = self.server.session
        self._probe = self.server.probe
        
        # Give the probe a session if it doesn't have one, in case it needs to access settings.
        # TODO: create a session proxy so client-side options can be accessed
        if self._probe.session is None:
            self._probe.session = self._session
        
        # Dict to store handles for AP memory interfaces.
        self._next_ap_memif_handle = 0
        self._ap_memif_handles = {}
    
        # Create the request handlers dict here so we can reference bound probe methods.
        self._REQUEST_HANDLERS = {
                # Command                Handler                            Arg count
                'hello':                (self._request__hello,              1   ),
                'readprop':             (self._request__read_property,      1   ),
                'open':                 (self._probe.open,                  0   ), # 'open'
                'close':                (self._probe.close,                 0   ), # 'close'
                'lock':                 (self._probe.lock,                  0   ), # 'lock'
                'unlock':               (self._probe.unlock,                0   ), # 'unlock'
                'connect':              (self._request__connect,            1   ), # 'connect', protocol:str
                'disconnect':           (self._probe.disconnect,            0   ), # 'disconnect'
                'swj_sequence':         (self._probe.swj_sequence,          2   ), # 'swj_sequence', length:int, bits:int
                'swd_sequence':         (self._probe.swd_sequence,          1   ), # 'swd_sequence', sequences:List[Union[Tuple[int], Tuple[int, int]]] -> Tuple[int, List[bytes]]
                'jtag_sequence':        (self._probe.jtag_sequence,         4   ), # 'jtag_sequence', cycles:int, tms:int, read_tdo:bool, tdi:int -> Union[None, int]
                'set_clock':            (self._probe.set_clock,             1   ), # 'set_clock', freq:int
                'reset':                (self._probe.reset,                 0   ), # 'reset'
                'assert_reset':         (self._probe.assert_reset,          1   ), # 'assert_reset', asserted:bool
                'is_reset_asserted':    (self._probe.is_reset_asserted,     0   ), # 'is_reset_asserted'
                'flush':                (self._probe.flush,                 0   ), # 'flush'
                'read_dp':              (self._probe.read_dp,               1   ), # 'read_dp', addr:int -> int
                'write_dp':             (self._probe.write_dp,              2   ), # 'write_dp', addr:int, data:int
                'read_ap':              (self._probe.read_ap,               1   ), # 'read_ap', addr:int -> int
                'write_ap':             (self._probe.write_ap,              2   ), # 'write_ap', addr:int, data:int
                'read_ap_multiple':     (self._probe.read_ap_multiple,      2   ), # 'read_ap_multiple', addr:int, count:int -> List[int]
                'write_ap_multiple':    (self._probe.write_ap_multiple,     2   ), # 'write_ap_multiple', addr:int, data:List[int]
                'get_memory_interface_for_ap': (self._request__get_memory_interface_for_ap, 2), # 'get_memory_interface_for_ap', ap_address_version:int, ap_nominal_address:int -> handle:int|null
                'swo_start':            (self._probe.swo_start,             1   ), # 'swo_start', baudrate:int
                'swo_stop':             (self._probe.swo_stop,              0   ), # 'swo_stop'
                'swo_read':             (self._request__swo_read,           0   ), # 'swo_read' -> List[int]
                'read_mem':             (self._request__read_mem,           3   ), # 'read_mem', handle:int, addr:int, xfer_size:int -> int
                'write_mem':            (self._request__write_mem,          4   ), # 'write_mem', handle:int, addr:int, value:int, xfer_size:int
                'read_block32':         (self._request__read_block32,       3   ), # 'read_block32', handle:int, addr:int, word_count:int -> List[int]
                'write_block32':        (self._request__write_block32,      3   ), # 'write_block32', handle:int, addr:int, data:List[int]
                'read_block8':          (self._request__read_block8,        3   ), # 'read_block8', handle:int, addr:int, word_count:int -> List[int]
                'write_block8':         (self._request__write_block8,       3   ), # 'write_block8', handle:int, addr:int, data:List[int]
            }
        
        # Let superclass do its thing. (Can't use super() here because the superclass isn't derived
        # from object in Py2.)
        StreamRequestHandler.setup(self)
    
    def finish(self):
        LOG.info("Remote probe client disconnected (%s from port %i)", self._client_domain, self.client_address[1])
        
        self._session = None
        StreamRequestHandler.finish(self)
    
    def _send_error_response(self, status=1, message=""):
        response_dict = {
                "id": self._current_request_id,
                "status": status,
                "error": message,
            }
        response = json.dumps(response_dict)
        TRACE.debug("response: %s", response)
        response_encoded = response.encode('utf-8')
        self.wfile.write(response_encoded + b"\n")
        
    def _send_response(self, result):
        response_dict = {
                "id": self._current_request_id,
                "status": 0,
            }
        if result is not None:
            response_dict["result"] = result
        response = json.dumps(response_dict)
        TRACE.debug("response: %s", response)
        response_encoded = response.encode('utf-8')
        self.wfile.write(response_encoded + b"\n")
        
    def handle(self):
        # Process requests until the connection is closed.
        while True:
            try:
                request = None
                request_dict = None
                self._current_request_id = -1
                
                # Read request line.
                request = self.rfile.readline()
                TRACE.debug("request: %s", request)
                if len(request) == 0:
                    LOG.debug("empty request, closing connection")
                    return
                
                try:
                    request_dict = json.loads(request)
                except json.JSONDecodeError:
                    self._send_error_response(message="invalid request format")
                    continue
            
                if not isinstance(request_dict, dict):
                    self._send_error_response(message="invalid request format")
                    continue
                    
                if 'id' not in request_dict:
                    self._send_error_response(message="missing request ID")
                    continue
                self._current_request_id = request_dict['id']
                
                if 'request' not in request_dict:
                    self._send_error_response(message="missing request field")
                    continue
                request_type = request_dict['request']
                
                # Get arguments. If the key isn't present then there are no arguments.
                request_args = request_dict.get('arguments', [])
            
                if not isinstance(request_args, list):
                    self._send_error_response(message="invalid request arguments format")
                    continue
                
                if request_type not in self._REQUEST_HANDLERS:
                    self._send_error_response(message="unknown request type")
                    continue
                handler, arg_count = self._REQUEST_HANDLERS[request_type]
                self._check_args(request_args, arg_count)
                result = handler(*request_args)
                
                # Send a success response.
                self._send_response(result)
            # Catch all exceptions so that an error response can be returned, to not leave the client hanging.
            except Exception as err:
                # Only send an error response if we received an request.
                if request is not None:
                    LOG.error("Error while processing %s request from client: %s", request, err,
                            exc_info=self._session.log_tracebacks)
                    self._send_error_response(status=self._get_exception_status_code(err),
                            message=str(err))
                else:
                    LOG.error("Error before request was received: %s", err,
                            exc_info=self._session.log_tracebacks)
                # Reraise non-pyocd errors.
                if not isinstance(err, exceptions.Error):
                    raise
    
    def _get_exception_status_code(self, err):
        """! @brief Convert an exception class into a status code."""
        # Must test the exception class in order of specific to general.
        if isinstance(err, exceptions.ProbeDisconnected):
            return self.StatusCode.PROBE_DISCONNECTED
        elif isinstance(err, exceptions.ProbeError):
            return self.StatusCode.PROBE_ERROR
        elif isinstance(err, exceptions.TransferFaultError):
            return self.StatusCode.TRANSFER_FAULT
        elif isinstance(err, exceptions.TransferTimeoutError):
            return self.StatusCode.TRANSFER_TIMEOUT
        elif isinstance(err, exceptions.TransferError):
            return self.StatusCode.TRANSFER_ERROR
        else:
            return self.StatusCode.GENERAL_ERROR
    
    def _check_args(self, args, count):
        if len(args) != count:
            raise exceptions.Error("malformed request; invalid number of arguments")
    
    def _request__hello(self, version):
        # 'hello', protocol-version:int
        if version != self.PROTOCOL_VERSION:
            raise exceptions.Error("client requested unsupported protocol version %i (expected %i)" %
                    (version, self.PROTOCOL_VERSION))
    
    def _request__read_property(self, name):
        # 'readprop', name:str
        if not hasattr(self._probe, name):
            raise exceptions.Error("unknown property name '%s' requested" % name)
        value = getattr(self._probe, name)
        # Run the property value through a value transformer if one is defined for this property.
        if name in self._PROPERTY_CONVERTERS:
            value = self._PROPERTY_CONVERTERS[name](value)
        return value
    
    def _request__connect(self, protocol_name):
        # 'connect', protocol:str
        try:
            protocol = DebugProbe.Protocol[protocol_name]
        except KeyError:
            raise exceptions.Error("invalid protocol name %s" % protocol_name)
        self._probe.connect(protocol)
    
    def _request__get_memory_interface_for_ap(self, ap_address_version, ap_nominal_address):
        # 'get_memory_interface_for_ap', ap_address_version:int, ap_nominal_address:int -> handle:int|null
        ap_version = APVersion(ap_address_version)
        if ap_version == APVersion.APv1:
            ap_address = APv1Address(ap_nominal_address)
        elif ap_version == APVersion.APv2:
            ap_address = APv2Address(ap_nominal_address)
        else:
            raise exceptions.Error("invalid AP version in remote get_memory_interface_for_ap request")

        memif = self._probe.get_memory_interface_for_ap(ap_address)
        if memif is not None:
            handle = self._next_ap_memif_handle
            self._next_ap_memif_handle += 1
            self._ap_memif_handles[handle] = memif
            LOG.debug("creating memif for AP%s (handle %i)", ap_address, handle)
        else:
            handle = None
        return handle
    
    def _request__swo_read(self):
        return list(self._probe.swo_read())

    def _request__read_mem(self, handle, addr, xfer_size):
        # 'read_mem', handle:int, addr:int, xfer_size:int -> int
        if handle not in self._ap_memif_handles:
            raise exceptions.Error("invalid handle received from remote memory access")
        return self._ap_memif_handles[handle].read_memory(addr, xfer_size, now=True)
    
    def _request__write_mem(self, handle, addr, value, xfer_size):
        # 'write_mem', handle:int, addr:int, value:int, xfer_size:int
        if handle not in self._ap_memif_handles:
            raise exceptions.Error("invalid handle received from remote memory access")
        self._ap_memif_handles[handle].write_memory(addr, value, xfer_size)
    
    def _request__read_block32(self, handle, addr, word_count):
        # 'read_block32', handle:int, addr:int, word_count:int -> List[int]
        # TODO use base64 data
        if handle not in self._ap_memif_handles:
            raise exceptions.Error("invalid handle received from remote memory access")
        return self._ap_memif_handles[handle].read_memory_block32(addr, word_count)
    
    def _request__write_block32(self, handle, addr, data):
        # 'write_block32', handle:int, addr:int, data:List[int]
        # TODO use base64 data
        if handle not in self._ap_memif_handles:
            raise exceptions.Error("invalid handle received from remote memory access")
        self._ap_memif_handles[handle].write_memory_block32(addr, data)
    
    def _request__read_block8(self, handle, addr, word_count):
        # 'read_block8', handle:int, addr:int, word_count:int -> List[int]
        # TODO use base64 data
        if handle not in self._ap_memif_handles:
            raise exceptions.Error("invalid handle received from remote memory access")
        return self._ap_memif_handles[handle].read_memory_block8(addr, word_count)
    
    def _request__write_block8(self, handle, addr, data):
        # 'write_block8', handle:int, addr:int, data:List[int]
        # TODO use base64 data
        if handle not in self._ap_memif_handles:
            raise exceptions.Error("invalid handle received from remote memory access")
        self._ap_memif_handles[handle].write_memory_block8(addr, data)

    _PROPERTY_CONVERTERS = {
            'capabilities':                 lambda value: [v.name for v in value],
            'supported_wire_protocols':     lambda value: [v.name for v in value],
            'wire_protocol':                lambda value: value.name if (value is not None) else None,
        }




# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import sleep
import logging
from typing import (Dict, Tuple)

from .debug_probe import DebugProbe
from ..core import exceptions
from ..core.plugin import Plugin
from .pydapaccess import DAPAccess
from ..board.mbed_board import MbedBoard
from ..board.board_ids import BOARD_ID_TO_INFO

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class CMSISDAPProbe(DebugProbe):
    """! @brief Wraps a pydapaccess link as a DebugProbe.
    
    Supports CMSIS-DAP v1 and v2.
    """

    # Masks for CMSIS-DAP capabilities.
    SWD_CAPABILITY_MASK = 1
    JTAG_CAPABILITY_MASK = 2

    # Map from DebugProbe protocol types to/from DAPAccess port types.
    #
    # Note that Protocol.DEFAULT gets mapped to PORT.SWD. We need a concrete port type because some
    # non-reference CMSIS-DAP implementations do not accept the default port type.
    PORT_MAP = {
        DebugProbe.Protocol.DEFAULT: DAPAccess.PORT.SWD,
        DebugProbe.Protocol.SWD: DAPAccess.PORT.SWD,
        DebugProbe.Protocol.JTAG: DAPAccess.PORT.JTAG,
        DAPAccess.PORT.DEFAULT: DebugProbe.Protocol.DEFAULT,
        DAPAccess.PORT.SWD: DebugProbe.Protocol.SWD,
        DAPAccess.PORT.JTAG: DebugProbe.Protocol.JTAG,
        }
    
    # APnDP constants.
    DP = 0
    AP = 1
    
    # Bitmasks for AP register address fields.
    A32 = 0x0000000c
    
    # Map from AP/DP and 2-bit register address to the enums used by pydapaccess.
    REG_ADDR_TO_ID_MAP: Dict[Tuple[int, int], DAPAccess.REG] = {
        # APnDP A32
        ( 0,    0x0 ) : DAPAccess.REG.DP_0x0,
        ( 0,    0x4 ) : DAPAccess.REG.DP_0x4,
        ( 0,    0x8 ) : DAPAccess.REG.DP_0x8,
        ( 0,    0xC ) : DAPAccess.REG.DP_0xC,
        ( 1,    0x0 ) : DAPAccess.REG.AP_0x0,
        ( 1,    0x4 ) : DAPAccess.REG.AP_0x4,
        ( 1,    0x8 ) : DAPAccess.REG.AP_0x8,
        ( 1,    0xC ) : DAPAccess.REG.AP_0xC,
        }
    
    ## USB VID and PID pair for DAPLink firmware.
    DAPLINK_VIDPID = (0x0d28, 0x0204)
    
    @classmethod
    def get_all_connected_probes(cls, unique_id=None, is_explicit=False):
        try:
            return [cls(dev) for dev in DAPAccess.get_connected_devices()]
        except DAPAccess.Error as exc:
            raise cls._convert_exception(exc) from exc
    
    @classmethod
    def get_probe_with_id(cls, unique_id, is_explicit=False):
        try:
            dap_access = DAPAccess.get_device(unique_id)
            if dap_access is not None:
                return cls(dap_access)
            else:
                return None
        except DAPAccess.Error as exc:
            raise cls._convert_exception(exc) from exc

    def __init__(self, device):
        super(CMSISDAPProbe, self).__init__()
        self._link = device
        self._supported_protocols = None
        self._protocol = None
        self._is_open = False
        self._caps = set()
    
    @property
    def board_id(self):
        """! @brief Unique identifier for the board.
        
        Only board IDs for DAPLink firmware are supported. We can't assume other
        CMSIS-DAP firmware is using the same serial number format, so we cannot reliably
        extract the board ID.
        
        @return Either a 4-character board ID string, or None if the probe doesn't have a board ID.
        """
        if self._link.vidpid == self.DAPLINK_VIDPID:
            return self.unique_id[0:4]
        else:
            return None

    @property
    def description(self):
        try:
            # self.board_id may be None.
            board_info = BOARD_ID_TO_INFO[self.board_id]
        except KeyError:
            return self.vendor_name + " " + self.product_name
        else:
            return "{0} [{1}]".format(board_info.name, board_info.target)
    
    @property
    def vendor_name(self):
        return self._link.vendor_name
    
    @property
    def product_name(self):
        return self._link.product_name

    @property
    def supported_wire_protocols(self):
        """! @brief Only valid after opening."""
        return self._supported_protocols

    @property
    def unique_id(self):
        return self._link.get_unique_id()

    @property
    def wire_protocol(self):
        return self._protocol
    
    @property
    def is_open(self):
        return self._is_open
    
    @property
    def capabilities(self):
        return self._caps

    def create_associated_board(self):
        assert self.session is not None
        
        # Only support associated Mbed boards for DAPLink firmware. We can't assume other
        # CMSIS-DAP firmware is using the same serial number format, so we cannot reliably
        # extract the board ID.
        if self.board_id is not None:
            return MbedBoard(self.session, board_id=self.board_id)
        else:
            return None
    
    def open(self):
        try:
            TRACE.debug("trace: open")
            
            self._link.open()
            self._is_open = True
            self._link.set_deferred_transfer(self.session.options.get('cmsis_dap.deferred_transfers'))
        
            # Read CMSIS-DAP capabilities
            self._capabilities = self._link.identify(DAPAccess.ID.CAPABILITIES)
            self._supported_protocols = [DebugProbe.Protocol.DEFAULT]
            if self._capabilities & self.SWD_CAPABILITY_MASK:
                self._supported_protocols.append(DebugProbe.Protocol.SWD)
            if self._capabilities & self.JTAG_CAPABILITY_MASK:
                self._supported_protocols.append(DebugProbe.Protocol.JTAG)
            
            self._caps = {
                self.Capability.SWJ_SEQUENCE,
                self.Capability.BANKED_DP_REGISTERS,
                self.Capability.APv2_ADDRESSES,
                self.Capability.JTAG_SEQUENCE,
                }
            if self._link.has_swd_sequence:
                self._caps.add(self.Capability.SWD_SEQUENCE)
            if self._link.has_swo():
                self._caps.add(self.Capability.SWO)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc
    
    def close(self):
        try:
            TRACE.debug("trace: close")
            
            self._link.close()
            self._is_open = False
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          Target control functions
    # ------------------------------------------- #
    def connect(self, protocol=None):
        TRACE.debug("trace: connect(%s)", protocol.name if (protocol is not None) else "None")
        
        # Convert protocol to port enum.
        # 
        # We must get a non-default port, since some CMSIS-DAP implementations do not accept the default
        # port. Note that the conversion of the default port type is contained in the PORT_MAP dict so it 
        # is one location.
        port = self.PORT_MAP.get(protocol, self.PORT_MAP[DebugProbe.Protocol.DEFAULT])
        assert port is not DAPAccess.PORT.DEFAULT
        
        try:
            self._link.connect(port)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc
        
        # Read the current mode and save it.
        actualMode = self._link.get_swj_mode()
        self._protocol = self.PORT_MAP[actualMode]

    def swj_sequence(self, length, bits):
        TRACE.debug("trace: swj_sequence(length=%i, bits=%x)", length, bits)
        
        try:
            self._link.swj_sequence(length, bits)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def swd_sequence(self, sequences):
        TRACE.debug("trace: swd_sequence(sequences=%r)", sequences)
        
        try:
            self._link.swd_sequence(sequences)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def jtag_sequence(self, cycles, tms, read_tdo, tdi):
        TRACE.debug("trace: jtag_sequence(cycles=%i, tms=%x, read_tdo=%s, tdi=%x)", cycles, tms, read_tdo, tdi)
        
        try:
            self._link.jtag_sequence(cycles, tms, read_tdo, tdi)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def disconnect(self):
        TRACE.debug("trace: disconnect")
        
        try:
            self._link.disconnect()
            self._protocol = None
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def set_clock(self, frequency):
        TRACE.debug("trace: set_clock(freq=%i)", frequency)
        
        try:
            self._link.set_clock(frequency)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def reset(self):
        TRACE.debug("trace: reset")
        
        try:
            self._link.assert_reset(True)
            sleep(self.session.options.get('reset.hold_time'))
            self._link.assert_reset(False)
            sleep(self.session.options.get('reset.post_delay'))
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def assert_reset(self, asserted):
        TRACE.debug("trace: assert_reset(%s)", asserted)
        
        try:
            self._link.assert_reset(asserted)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc
    
    def is_reset_asserted(self):
        try:
            result = self._link.is_reset_asserted()
            TRACE.debug("trace: is_reset_asserted -> %s", result)
            return result
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def flush(self):
        TRACE.debug("trace: flush")
        
        try:
            self._link.flush()
        except DAPAccess.Error as exc:
            TRACE.debug("trace: error from flush: %r", exc)
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          DAP Access functions
    # ------------------------------------------- #

    def read_dp(self, addr, now=True):
        reg_id = self.REG_ADDR_TO_ID_MAP[self.DP, addr]
        
        try:
            if not now:
                TRACE.debug("trace: read_dp(addr=%#010x) -> ...", addr)
            result = self._link.read_reg(reg_id, now=now)
        except DAPAccess.Error as error:
            TRACE.debug("trace: read_dp(addr=%#010x) -> error(%s)", addr, error)
            raise self._convert_exception(error) from error

        # Read callback returned for async reads.
        def read_dp_result_callback():
            try:
                value = result()
                TRACE.debug("trace: ... read_dp(addr=%#010x) -> %#010x", addr, value)
                return value
            except DAPAccess.Error as error:
                TRACE.debug("trace: ... read_dp(addr=%#010x) -> error(%s)", addr, error)
                raise self._convert_exception(error) from error

        if now:
            TRACE.debug("trace: read_dp(addr=%#010x) -> %#010x", addr, result)
            return result
        else:
            return read_dp_result_callback

    def write_dp(self, addr, data):
        reg_id = self.REG_ADDR_TO_ID_MAP[self.DP, addr]
        
        # Write the DP register.
        try:
            self._link.write_reg(reg_id, data)
            TRACE.debug("trace: write_dp(addr=%#010x, data=%#010x)", addr, data)
        except DAPAccess.Error as error:
            TRACE.debug("trace: write_dp(addr=%#010x, data=%#010x) -> error(%s)", addr, data, error)
            raise self._convert_exception(error) from error

        return True

    def read_ap(self, addr, now=True):
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]

        try:
            if not now:
                TRACE.debug("trace: read_ap(addr=%#010x) -> ...", addr)
            result = self._link.read_reg(ap_reg, now=now)
        except DAPAccess.Error as error:
            raise self._convert_exception(error) from error

        # Read callback returned for async reads.
        def read_ap_result_callback():
            try:
                value = result()
                TRACE.debug("trace: ... read_ap(addr=%#010x) -> %#010x", addr, value)
                return value
            except DAPAccess.Error as error:
                TRACE.debug("trace: ... read_ap(addr=%#010x) -> error(%s)", addr, error)
                raise self._convert_exception(error) from error

        if now:
            TRACE.debug("trace: read_ap(addr=%#010x) -> %#010x", addr, result)
            return result
        else:
            return read_ap_result_callback

    def write_ap(self, addr, data):
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]

        try:
            # Perform the AP register write.
            self._link.write_reg(ap_reg, data)
            TRACE.debug("trace: write_ap(addr=%#010x, data=%#010x)", addr, data)
        except DAPAccess.Error as error:
            TRACE.debug("trace: write_ap(addr=%#010x, data=%#010x) -> error(%s)", addr, data, error)
            raise self._convert_exception(error) from error

        return True

    def read_ap_multiple(self, addr, count=1, now=True):
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]
        
        try:
            if not now:
                TRACE.debug("trace: read_ap_multi(addr=%#010x, count=%i) -> ...", addr, count)
            result = self._link.reg_read_repeat(count, ap_reg, dap_index=0, now=now)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

        # Need to wrap the deferred callback to convert exceptions.
        def read_ap_repeat_callback():
            try:
                values = result()
                TRACE.debug("trace: ... read_ap_multi(addr=%#010x, count=%i) -> [%s]", addr, count,
                        ", ".join(["%#010x" % v for v in values]))
                return values
            except DAPAccess.Error as exc:
                TRACE.debug("trace: ... read_ap_multi(addr=%#010x, count=%i) -> error(%s)",
                    addr, count, exc)
                raise self._convert_exception(exc) from exc

        if now:
            TRACE.debug("trace: read_ap_multi(addr=%#010x, count=%i) -> [%s]", addr, count,
                    ", ".join(["%#010x" % v for v in result]))
            return result
        else:
            return read_ap_repeat_callback

    def write_ap_multiple(self, addr, values):
        assert isinstance(addr, int)
        ap_reg = self.REG_ADDR_TO_ID_MAP[self.AP, (addr & self.A32)]
        
        try:
            self._link.reg_write_repeat(len(values), ap_reg, values, dap_index=0)
            TRACE.debug("trace: write_ap_multi(addr=%#010x, (%i)[%s])", addr, len(values),
                    ", ".join(["%#010x" % v for v in values]))
        except DAPAccess.Error as exc:
            TRACE.debug("trace: write_ap_multi(addr=%#010x, (%i)[%s]) -> error(%s)", addr, len(values),
                    ", ".join(["%#010x" % v for v in values]), exc)
            raise self._convert_exception(exc) from exc
    
    # ------------------------------------------- #
    #          SWO functions
    # ------------------------------------------- #

    def swo_start(self, baudrate):
        TRACE.debug("trace: swo_start(baud=%i)", baudrate)
        
        try:
            self._link.swo_configure(True, baudrate)
            self._link.swo_control(True)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def swo_stop(self):
        TRACE.debug("trace: swo_stop")
        
        try:
            self._link.swo_configure(False, 0)
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    def swo_read(self):
        try:
            data = self._link.swo_read()
            TRACE.debug("trace: swo_read -> %i bytes", len(data))
            return data
        except DAPAccess.Error as exc:
            raise self._convert_exception(exc) from exc

    @staticmethod
    def _convert_exception(exc):
        if isinstance(exc, DAPAccess.TransferFaultError):
            return exceptions.TransferFaultError(*exc.args)
        elif isinstance(exc, DAPAccess.TransferTimeoutError):
            return exceptions.TransferTimeoutError(*exc.args)
        elif isinstance(exc, DAPAccess.TransferError):
            return exceptions.TransferError(*exc.args)
        elif isinstance(exc, (DAPAccess.DeviceError, DAPAccess.CommandError)):
            return exceptions.ProbeError(*exc.args)
        elif isinstance(exc, DAPAccess.Error):
            return exceptions.Error(*exc.args)
        else:
            return exc

class CMSISDAPProbePlugin(Plugin):
    """! @brief Plugin class for CMSISDAPProbe."""
    
    def load(self):
        return CMSISDAPProbe
    
    @property
    def name(self):
        return "cmsisdap"
    
    @property
    def description(self):
        return "CMSIS-DAP debug probe"



# pyOCD debugger
# Copyright (c) 2016-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging
from unittest import mock

from .mockcore import MockCore

@pytest.fixture(scope='function')
def mockcore():
    return MockCore()

@pytest.fixture(scope='function')
def mockcore_no_fpu():
    return MockCore(has_fpu=False)

# Ignore any tests that are currently causing issues.
collect_ignore = [
    # None right now.
    ]



# pyOCD debugger
# Copyright (c) 2021 Huada Semiconductor Corporation
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile


#DEBUG_ACTIVE
DEBUG_ACTIVE = 0x40002038
DEBUG_ACTIVE_VAL = 0x00000FFF

FLASH_ALGO = { 
    'load_address' : 0x20000000,
    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x4770ba40, 0x4770bac0, 0x4c5fb570, 0x06806820, 0xd0210f85, 0xf8b2f000, 0x21016820, 0x60204308,
    0x050068e0, 0xf000d5fc, 0x6820f8a9, 0x43882130, 0xf0006020, 0x2d01f8a3, 0x2d02d004, 0x2d03d005,
    0xe005d10a, 0x21026820, 0x6820e004, 0xe0012104, 0x21086820, 0x60204388, 0xf890f000, 0x21ff6820,
    0x438831c1, 0xf0006020, 0x6820f889, 0x026d2503, 0x602043a8, 0xf882f000, 0x43a868e0, 0xf00060e0,
    0x4c42f877, 0x60202020, 0xf872f000, 0x60602017, 0xf86ef000, 0x60a0201b, 0xf86af000, 0x60e0483c,
    0xf866f000, 0x6120483b, 0xf862f000, 0x61602018, 0xf85ef000, 0x61a020f0, 0xf85af000, 0x00c0207d,
    0x200061e0, 0x2000bd70, 0xb5704770, 0x6a204c2f, 0xd4fc06c0, 0xf84cf000, 0x6320482f, 0xf848f000,
    0x21036a20, 0x62204308, 0x602d2500, 0x06c06a20, 0xf000d4fc, 0x6325f83d, 0xbd702000, 0x4c23b570,
    0x6a214605, 0xd4fc06c8, 0xf832f000, 0x63204822, 0xf82ef000, 0x08806a20, 0x1c800080, 0x26006220,
    0x6a20602e, 0xd4fc06c0, 0xf822f000, 0x20006326, 0xb5f7bd70, 0x460f4616, 0xf0002400, 0x4d13f819,
    0x63284815, 0xf814f000, 0x08806a28, 0x1c400080, 0xe0066228, 0x5d319800, 0x6a285501, 0xd4fc06c0,
    0x42bc1c64, 0xf000d3f6, 0x2000f803, 0xbdfe6328, 0x490a4806, 0x490a62c1, 0x477062c1, 0x49074802,
    0x49076081, 0x47706081, 0x40002000, 0x40020000, 0x00004650, 0x000222e0, 0x0000ffff, 0x00005a5a,
    0x0000a5a5, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000029,
    'pc_unInit': 0x200000e7,
    'pc_program_page': 0x20000153,
    'pc_erase_sector': 0x2000011d,
    'pc_eraseAll': 0x200000eb,

    'static_base' : 0x20000000 + 0x00000020 + 0x000001a4,
    'begin_stack' : 0x20000400,
    'begin_data' : 0x20000000 + 0x600,
    'page_size' : 0x200,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20000600],   
    'min_program_length' : 0x200,
  }


class HC32L110(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x8000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x1000)
        )

    def __init__(self, session):
        super(HC32L110, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L110.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32F003(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x4000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x800)
        )

    def __init__(self, session):
        super(HC32F003, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32F003.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32F005(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x8000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x1000)
        )

    def __init__(self, session):
        super(HC32F005, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32F005.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)



# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

# Note
#  To run this script GNU Tools ARM Embedded must be installed,
#  along with python for the same architecture.  The program
#  "arm-none-eabi-gdb-py.exe" requires python for the same
#  architecture (x86 or 64) to work correctly. Also, on windows
#  the GNU Tools ARM Embedded bin directory needs to be added to
#  your path.

import os
import json
import sys
from subprocess import (
    Popen,
    STDOUT,
    PIPE,
    check_output,
    )
import argparse
import logging
import traceback
import threading

from pyocd.__main__ import PyOCDTool
from pyocd.core.helpers import ConnectHelper
from pyocd.utility.compatibility import to_str_safe
from pyocd.core.memory_map import MemoryType
from pyocd.flash.file_programmer import FileProgrammer
from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    binary_to_elf_file,
    get_env_file_name,
    get_test_binary_path,
    TEST_DIR,
    TEST_OUTPUT_DIR,
    ensure_output_dir,
    wait_with_deadline,
    )

# TODO, c1728p9 - run script several times with
#       with different command line parameters

LOG = logging.getLogger(__name__)

PYTHON_GDB = "arm-none-eabi-gdb-py"
TEST_TIMEOUT_SECONDS = 60.0 * 5

GDB_SCRIPT_PATH = os.path.join(TEST_DIR, "gdb_test_script.py")

class GdbTestResult(TestResult):
    def __init__(self):
        super(self.__class__, self).__init__(None, None, None)
        self.name = "gdbserver"


class GdbTest(Test):
    def __init__(self):
        super(self.__class__, self).__init__("Gdb Test", test_gdb)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id, self.n)
        except Exception as e:
            result = GdbTestResult()
            result.passed = False
            print("Exception %s when testing board %s" %
                  (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

TEST_RESULT_KEYS = [
    "breakpoint_count",
    "watchpoint_count",
    "step_time_si",
    "step_time_s",
    "step_time_n",
    "fail_count",
]


def test_gdb(board_id=None, n=0):
    temp_test_elf_name = None
    result = GdbTestResult()
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        memory_map = board.target.get_memory_map()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        rom_region = memory_map.get_boot_memory()
        target_type = board.target_type
        binary_file = get_test_binary_path(board.test_binary)
        if board_id is None:
            board_id = board.unique_id
        target_test_params = get_target_test_params(session)
        test_port = 3333 + n
        telnet_port = 4444 + n
        
        # Hardware breakpoints are not supported above 0x20000000 on
        # Cortex-M devices with FPB revision 1.
        fpb = session.target.selected_core.fpb
        assert fpb is not None
        ignore_hw_bkpt_result = int(fpb.revision == 1 and ram_region.start >= 0x20000000)

        # Program with initial test image
        FileProgrammer(session).program(binary_file, base_address=rom_region.start)

    # Generate an elf from the binary test file.
    temp_test_elf_name = binary_to_elf_file(binary_file, rom_region.start)

    # Write out the test configuration
    test_params = {
        "test_port" : test_port,
        "rom_start" : rom_region.start,
        "rom_length" : rom_region.length,
        "ram_start" : ram_region.start,
        "ram_length" : ram_region.length,
        "invalid_start" : 0x3E000000,
        "invalid_length" : 0x1000,
        "expect_error_on_invalid_access" : target_test_params['error_on_invalid_access'],
        "ignore_hw_bkpt_result" : ignore_hw_bkpt_result,
        "test_elf" : temp_test_elf_name,
        }
    test_param_filename = os.path.join(TEST_OUTPUT_DIR, "gdb_test_params%s_%d.txt" % (get_env_file_name(), n))
    with open(test_param_filename, "w") as f:
        f.write(json.dumps(test_params))

    # Remove result from previous run.
    test_result_filename = os.path.join(TEST_OUTPUT_DIR, "gdb_test_results%s_%d.txt" % (get_env_file_name(), n))
    if os.path.exists(test_result_filename):
        os.remove(test_result_filename)

    # Run the test
    gdb_args = [PYTHON_GDB, "--nh", "-ex", "set $testn=%d" % n, "--command=%s" % GDB_SCRIPT_PATH]
    gdb_output_filename = os.path.join(TEST_OUTPUT_DIR, "gdb_output%s_%s_%d.txt" % (get_env_file_name(), board.target_type, n))
    with open(gdb_output_filename, "w") as f:
        LOG.info('Starting gdb (stdout -> %s): %s', gdb_output_filename, ' '.join(gdb_args))
        gdb_program = Popen(gdb_args, stdin=PIPE, stdout=f, stderr=STDOUT)
        server_args = ['gdbserver',
                '--port=%i' % test_port,
                "--telnet-port=%i" % telnet_port,
                "--frequency=%i" % target_test_params['test_clock'],
                "--uid=%s" % board_id,
                ]
        server = PyOCDTool()
        LOG.info('Starting gdbserver: %s', ' '.join(server_args))
        server_thread = threading.Thread(target=server.run, args=[server_args])
        server_thread.daemon = True
        server_thread.start()
        LOG.info('Waiting for gdb to finish...')
        did_complete = wait_with_deadline(gdb_program, TEST_TIMEOUT_SECONDS)
        LOG.info('Waiting for server to finish...')
        server_thread.join(timeout=TEST_TIMEOUT_SECONDS)
        if not did_complete:
            LOG.error("Test timed out!")
        if server_thread.is_alive():
            LOG.error('Server is still running!')

    try:
        with open(gdb_output_filename, 'r') as f:
            LOG.debug('Gdb output:\n%s', f.read())
    except IOError:
        pass

    # Read back the result
    result.passed = False
    if did_complete:
        try:
            with open(test_result_filename, "r") as f:
                test_result = json.loads(f.read())

            # Print results
            if set(TEST_RESULT_KEYS).issubset(test_result):
                print("----------------Test Results----------------")
                print("HW breakpoint count: %s" % test_result["breakpoint_count"])
                print("Watchpoint count: %s" % test_result["watchpoint_count"])
                print("Average instruction step time: %s" %
                      test_result["step_time_si"])
                print("Average single step time: %s" % test_result["step_time_s"])
                print("Average over step time: %s" % test_result["step_time_n"])
                print("Failure count: %i" % test_result["fail_count"])
                result.passed = test_result["fail_count"] == 0
        except IOError as err:
            LOG.error("Error reading test results: %s", err, exc_info=True)

    if result.passed:
        print("GDB TEST PASSED")
    else:
        print("GDB TEST FAILED")

    # Cleanup
    try:
        if temp_test_elf_name and os.path.exists(temp_test_elf_name):
            os.remove(temp_test_elf_name)
        os.remove(test_result_filename)
        os.remove(test_param_filename)
    except IOError as err:
        pass

    return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD gdb test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    ensure_output_dir()
    test_gdb()



# pyOCD debugger
# Copyright (c) 2021 NXP
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile

FLASH_ALGO = {
    'load_address' : 0x10000200,

    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x47700a80, 0x48474948, 0x60084449, 0x22004847, 0x21016102, 0x61426141, 0x61816141, 0x20024943,
    0x70083940, 0x47704610, 0x47702000, 0x4c40b5f8, 0x444c2032, 0x25004621, 0xc461263f, 0x483c3114,
    0x44484f3c, 0x91003c0c, 0x696047b8, 0xd10d2800, 0xc4612034, 0x44484834, 0x60206800, 0x3c0c4834,
    0x99004448, 0x696047b8, 0xd0002800, 0xbdf82001, 0x4d2fb5f8, 0x20320a84, 0xc511444d, 0x310c4629,
    0x4e2c482b, 0x460f602c, 0x3d084448, 0x696847b0, 0xd10e2800, 0xc5112034, 0x602c4823, 0x68004448,
    0x48236068, 0x44484639, 0x47b03d08, 0x28006968, 0x2001d000, 0xb5f8bdf8, 0x00064614, 0xcc03d10e,
    0x18406862, 0x18896821, 0x68a11840, 0x68e11840, 0x69211840, 0x42401840, 0x3c086160, 0x0ab04d14,
    0x2132444d, 0x60296068, 0x462960a8, 0x4f113114, 0x91004628, 0x696847b8, 0xd1112800, 0x60ac2033,
    0x20ffc541, 0x60683001, 0x44484807, 0x60a86800, 0x3d084807, 0x99004448, 0x696847b8, 0xd0002800,
    0xbdf82001, 0x00002ee0, 0x00000004, 0x40048040, 0x00000008, 0x0f001ff1, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x10000225,
    'pc_unInit': 0x10000249,
    'pc_program_page': 0x100002d7,
    'pc_erase_sector': 0x10000291,
    'pc_eraseAll': 0x1000024d,

    'begin_data' : 0x10000380,
    'begin_stack' : 0x10000200,
    'static_base' : 0x10000358,

    # 'page_size' : 0x100,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x10001000, 0x10001100],   # Enable double buffering
    'min_program_length' : 0x100,
}

class LPC845(CoreSightTarget):

    VENDOR = "NXP"

    MEMORY_MAP = MemoryMap(
        FlashRegion(    start=0,           length=0x10000,       blocksize=0x400, page_size=0x100, is_boot_memory=True, algo=FLASH_ALGO),
        RamRegion(      start=0x10000000,  length=0x3fe0)
        )

    def __init__(self, session):
        super(LPC845, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("LPC845.xml")

    def reset_and_halt(self, reset_type=None, map_to_user=True):
        super(LPC845, self).reset_and_halt(reset_type)

        # Remap to use flash and set SP and SP accordingly
        if map_to_user:
            self.write_memory(0x40048000, 0x2)
            sp = self.read_memory(0x0)
            pc = self.read_memory(0x4)
            self.write_core_register('sp', sp)
            self.write_core_register('pc', pc)
